#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it 
# and/or modify it under the terms of the GNU General Public 
# License as published by the Free Software Foundation; either 
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be 
# useful, but WITHOUT ANY WARRANTY; without even the implied 
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public 
# License along with program; see the file COPYING. If not, 
# write to the Free Software Foundation, Inc., 59 Temple Place 
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GBaseApp.py
#
# DESCRIPTION:
# Class that provides a basis for GNUe applications.
#
# NOTES:
# Typically, this class will not be called; rather, a tool will
# be a GClientApp or GServerApp.
#
# HISTORY:
#

import GConfig
#
# i18n support
#
import getopt
import sys
import string
import os
import ConfigParser
import gettext
from gnue.common import lineWrap

try:
  gettext.install('gnue',GConfig.getInstalledBase('install_prefix')+'/translations')
except IOError:
#  sys.stderr.write('Unable to load locale information falling back to default English language translations.\n')
  import __builtin__
  __builtin__.__dict__['_'] = lambda string:string


import GDebug
import GConnections

class GBaseApp:

  # Attributes to be overwritten by subclasses
  VERSION = "0.0.0"
  NAME = "GNUe Application"
  COMMAND_OPTIONS = []  # Should be in same format as _base_options below
  SUMMARY = "A brief summary of the program goes here."
  COMMAND = "app"
  USAGE = "[options]"


  # More options, but won't be changed unless
  # this is a non-GNUe app using GNUe-Common
  AUTHOR = "GNU Enterprise Project"
  EMAIL = "info@gnue.org"
  REPORT_BUGS_TO = "Please report any bugs to info@gnue.org."


  #  Run the program
  #  Should be overwritten by subclasses
  def run(self):
    pass


  # Attributes that will be set by GClientApp after __init__ has run
  OPTIONS = {}        # Will contain a hash containing command line options
  ARGUMENTS = []      # Will contain an array of command line arguments
  connections = None  # Will contain a GConnection object




  def __init__(self, connections=None, application=None, defaults=None):


    # format of COMMAND_OPTIONS and _base_options:
    #  [option, option] where option is:
    #       [dictionary key name, short option, long option,
    #        argument follows option?, argument name, help message]
    self._base_options = [
         [ 'version',None,'version',0,0, None,
           _('Displays the version information for this program.') ],
         [ 'debug_level',None,'debug-level',1,0, "level",
           _('Enables debugging messages.  Argument specifies the ') + \
           _('level of messages to display (e.g., "--debug-level 5" displays ') + \
           _('all debugging messages at level 5 or below.)') ],
         [ 'debug_file',None,'debug-file',1,None, "file",
           _('Sends all debugging messages to a specified file ') + \
           _('(e.g., "--debug-file trace.log" sends all output to "trace.log")') ],
         [ 'connections', None, 'connections', 1, None, "loc",
           _('Specifies the location of the connection definition file. ') + \
           _('<loc> may specify a file name ') + \
           _('(/usr/local/gnue/etc/connections.conf), ') + \
           _('or a URL location ') + \
           _('(http://localhost/connections.conf). ')+ \
           _('If this option is not specified, the environent variable ') + \
           _('GNUE_CONNECTIONS is checked.') ],
         [ 'help', None, 'help', 0, None, None,
           _('Displays this help screen.') ],
         [ 'configuration_options', None, 'configuration-options', 0, None, None,
           _('Displays a list of valid configuration file entries, their purpose, and their default values.') ],
         [ 'man_page', None, 'generate-man-page', 0, None, None,
           _('Generates a groff-formatted man page as a file in the current directory.') ],
         [ 'profile', None, 'profile', 0, None, None,
           _("Run Python's built-in profiler and display the resulting ") + \
           _("run statistics.") ],
         [ 'debugger', None, 'interactive-debugger', 0, None, None,
           _("Run the app inside Python's built-in debugger ")],
    ]

    # Python version check
    try:
      if sys.hexversion < 0x02000000:
        self.handleStartupError (_('This application requires Python 2.0 or greater.  You are running Python %s') % sys.version[:5])
    except AttributeError:
      # Really, really old version of Python...
      self.handleStartupError (_('This application requires Python 2.0 or greater.'))


    #
    # Get all command line options and arguments
    #
    shortoptions = ""
    longoptions = []
    lookup = {}
    for optionset in [self._base_options, self.COMMAND_OPTIONS]:
      for option in optionset:
        self.OPTIONS[option[0]] = option[4]
        if option[1] != None: 
          shortoptions = shortoptions + option[1]
          lookup["-" + option[1]] = option[0]
        lookup["--" + option[2]] = option[0]
        lo = option[2]
        if option[3]:   
          lo = lo + '='
          shortoptions = shortoptions + ':'
        longoptions.append(lo)

    try:
      opt, self.ARGUMENTS = getopt.getopt(sys.argv[1:], shortoptions, longoptions)
    except getopt.error, msg:
      self.handleStartupError(msg)

    for o in opt:
      if len(o[1]):
        self.OPTIONS[lookup[o[0]]] = o[1]
      else:
        self.OPTIONS[lookup[o[0]]] = 1

    if self.OPTIONS['help']:
      self.printHelp()
      sys.exit()
    elif self.OPTIONS['version']:
      self.printVersion()
      sys.exit()
    elif self.OPTIONS['man_page']:
      self.dumpManPage()
      sys.exit()
    elif self.OPTIONS['configuration_options']:
      print GConfig.printableConfigOptions(defaults)
      sys.exit()

    self._run = self.run

    # Should we profile?
    if self.OPTIONS['profile']:
      self.run = self._profile

    # Setup debugging
    # Should we run in debugger?
    if self.OPTIONS['debugger']:
      self.run = self._debugger

    try:
      GDebug.setDebug(int("%s" % self.OPTIONS['debug_level']),
          self.OPTIONS['debug_file'])
    except ValueError:
      self.handleStartupError(_('The debug_level option ("-d") expects a numerical value.'))

        
    
    GDebug.printMesg(1,"Python %s" % sys.version)
    GDebug.printMesg(1,"Run Options: %s" % opt)
    GDebug.printMesg(1,"Run Arguments: %s" % self.ARGUMENTS)

    # Read the config files
    if application:
      try:
        self.configurationManager = GConfig.GConfig(application, defaults)
      except ConfigParser.NoSectionError, msg:
        self.handleStartupError(_('The gnue.conf file is incomplete: ')\
                                + '\n   %s'         % msg)
      except Exception, msg:
        type=string.splitfields(str(sys.exc_type),'.')
        type.reverse()
        self.handleStartupError( \
          _('%s while reading gnue.conf: ') % type[0] \
          + '\n   %s'         % msg)

    # Get the connection definitions
    if connections != None:
      GDebug.printMesg(1,"Reusing connections instance")
      self.connections = connections
    else:
      if self.OPTIONS['connections']:
        self.connections_file = self.OPTIONS['connections']
      elif os.environ.has_key('GNUE_CONNECTIONS'):
        self.connections_file = os.environ['GNUE_CONNECTIONS']
      else:
        self.connections_file = "" # temp
#        self.handleStartupError(
#            _('Unable to load the connections definition file.\n') \
#            + _('\n   Please set the environmental variable GNUE_CONNECTIONS or ') \
#            + _('\n   use the "-f" command option.'))

      GDebug.printMesg(1, 'Connection Definition: "%s"' % self.connections_file)

      try:
        self.connections = GConnections.GConnections(self.connections_file)
      except GConnections.InvalidFormatError, msg:
        self.handleStartupError(
            _('Unable to load the connections definition file.\n') \
            + _('\n   The connections file is in an invalid format. ') \
            + _('\n   %s') \
               % msg)
      except IOError:
        self.handleStartupError(
             _('Unable to load the connections definition file.\n') \
             + _('\n   The connections file specified either does ') \
             + _('\n   not exist or is not readable by your account.\n') \
             + _('\n   Location: "%s"')
                % self.connections_file)

  #
  #  Display version information for this application
  #
  def printVersion(self):
    from gnue.common import VERSION as commonVersion
    print _("\n%s\nVersion %s\n") % (self.NAME, self.VERSION)
    print _("GNUe Common Version %s\n") % commonVersion

  #
  #  Display help information for this program
  #
  def printHelp(self):
    allOptions = {}
    descriptors = {}
    maxLength = 0

    for optionset in [self._base_options, self.COMMAND_OPTIONS]:
      for option in optionset:
        allOptions[string.upper(option[2])] = option

        if option[5]:
          descr = '--%s <%s>' % (option[2], option[5])
        else:
          descr = '--%s' % (option[2])
        if option[1]:
          descr += ', -%s' % option[1]

        descriptors[string.upper(option[2])] = descr

        if len(descr) > maxLength:
          maxLength = len(descr)

    if maxLength > 20:
      maxLength = 20

    sorted = allOptions.keys()
    sorted.sort()

    dispOptions = ""

    for optionKey in sorted:

      margin = maxLength + 4
      width = 78 - margin
      pos = 0

      if len(descriptors[optionKey]) > maxLength:
        dispOptions += "\n  %s\n%s" % (descriptors[optionKey], " " * margin)
      else:
        dispOptions += "\n  %s  %s" % (descriptors[optionKey],
               " " * (maxLength - len(descriptors[optionKey])))

      for word in string.split(allOptions[optionKey][6]):
        if (len(word) + pos) > width:
          pos = 0
          dispOptions = dispOptions + "\n" + " " * margin

        pos = pos + len(word) + 1

        dispOptions = dispOptions + word + " "

      dispOptions = dispOptions + "\n"

    self.printVersion()
    print _("Usage:  %s %s\n\n%s\n\nAvailable command line options:\n%s") % \
       (self.COMMAND, self.USAGE, self.SUMMARY, dispOptions)
    print _("%s\n" % self.REPORT_BUGS_TO)


  #
  #  Create a man page for this utility.
  #
  def dumpManPage(self):
    import time
    year = time.strftime("%Y", time.localtime(time.time()))


    dest = open("%s.1" % self.COMMAND,"w")

    COMMAND = string.replace(self.COMMAND,'-','\-')

    #
    # HEADER + NAME Sections
    #
    dest.write ("""\
.ig
Copyright (C) 2000-%s Free Software Foundation, Inc.

Permission is granted to make and distribute verbatim copies of
this manual provided the copyright notice and this permission notice
are preserved on all copies.

Permission is granted to copy and distribute modified versions of this
manual under the conditions for verbatim copying, provided that the
entire resulting derived work is distributed under the terms of a
permission notice identical to this one.

Permission is granted to copy and distribute translations of this
manual into another language, under the above conditions for modified
versions, except that this permission notice may be included in
translations approved by the Free Software Foundation instead of in
the original English.
..
""" % year)

    dest.write('.TH %s 1 "%s" "%s"\n' % (
            string.upper(COMMAND),
            time.strftime("%d %B %Y",
                          time.localtime(time.time())),
            self.NAME))
    dest.write('.SH NAME\n')
    dest.write('%s \- %s\n' % (COMMAND,
                               string.replace(self.NAME,'-','\-')))


    #
    # SYNOPSIS Section
    #
    dest.write('.SH SYNOPSIS\n')
    dest.write('.ll +8\n')
    dest.write('.B %s\n' % COMMAND)
    for p in string.split(self.USAGE):
      part = p
      if part[0] == '[' and part[-1] == ']':
        pre = '[\n.I '
        post = '\n]'
        part = part[1:-1]
      else:
        pre = '.I '
        post = ''

      if part == '...':
        part = '\&...'

      if string.find(part,'=') > 0:
        first, last = string.split(part,'=',2)
        part = string.replace(part,
             '=','\n=\n.I ',1)

      dest.write('%s%s%s\n' % (pre, part, post))


    #
    # DESCRIPTION Section
    #
    dest.write('.SH DESCRIPTION\n')
    dest.write(lineWrap(string.replace(self.SUMMARY,'-','\-'),70) + '\n')


    #
    # OPTIONS Section
    #
    dest.write('.SH OPTIONS\n')

    allOptions = {}
    descriptors = {}
    for optionset in [self._base_options, self.COMMAND_OPTIONS]:
      for option in optionset:
        allOptions[string.upper(option[2])] = option

        if option[5]:
          descr = '.B \-\-%s <%s>' % (option[2], option[5])
        else:
          descr = '.B \-\-%s' % (option[2])
        if option[1]:
          descr += ', \-%s' % option[1]

        descriptors[string.upper(option[2])] = descr


    sorted = allOptions.keys()
    sorted.sort()

    dispOptions = ""

    for optionKey in sorted:
      dest.write(".TP\n%s\n" % (descriptors[optionKey]))
      dest.write(string.replace(string.replace(
               lineWrap(
               string.replace(allOptions[optionKey][6],'-','\-'),70),
                "\n.", "\n\\."), "\n'", "\n\\'"))

    #
    # AUTHOR, BUGS, and COPYRIGHT sections
    #
    dest.write("""\
.SH AUTHOR
%(AUTHOR)s <%(EMAIL)s>
.SH BUGS
%(BUGS)s
Include a complete, self-contained example
that will allow the bug to be reproduced,
and say which version of this tool you are using.
.SH COPYRIGHT
Copyright \(co 2000-%(YEAR)s Free Software Foundation, Inc.
.LP
%(COMMAND)s is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.
.LP
%(COMMAND)s is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.
.LP
You should have received a copy of the GNU General Public License along
with %(COMMAND)s; see the file COPYING.  If not, write to the Free Software
Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
""" % {'COMMAND': COMMAND,
       'YEAR': year,
       'BUGS': self.REPORT_BUGS_TO,
       'AUTHOR': self.AUTHOR,
       'EMAIL': self.EMAIL})

    #
    # Cleanup
    #
    dest.close()
    print "Man file create: %s.1" % self.COMMAND


  #
  #  Display a startup error and exit gracefully with a message on
  #  how to get help
  #
  def handleStartupError(self, msg):
    self.printVersion()

    # if msg is multiline, then surround with dashes to set it apart
    if string.find("%s" % msg, "\n") + 1:
      print '-' * 60
    print _("Error: %s") % msg
    if string.find("%s" % msg, "\n") + 1:
      print '-' * 60
    print _("\nFor help, type:\n   %s --help\n") % (self.COMMAND)
    sys.exit()


  #
  # Used when interactive debugger in use
  #
  def _debugger(self):

    import pdb
    debugger = pdb.Pdb()
    GDebug.setDebugger(debugger)
    debugger.runctx( 'self._run()', globals(), locals() )

  #
  #  Used when profiling
  #
  def _profile(self):

    import profile
    prof = profile.Profile()
    prof.runctx( 'self._run()', globals(), locals() )

    import pstats
    p = pstats.Stats(prof)
    p.sort_stats('time').print_stats(50)
    p.sort_stats('cumulative').print_stats(50)
    p.sort_stats('calls').print_stats(50)



