#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GConnections.py
#
# DESCRIPTION:
# Class that loads connection definition files and maintains
# database connections.
#
# NOTES:
#
# HISTORY:
#

from ConfigParser import *
import sys, string, copy, netrc
import GDebug
import GDataObjects
import GLoginHandler
from gnue.common import openResource, dyn_import

class Error(StandardError):
  # Base error
  pass

class NotFoundError (Error):
  # Raised if a requested connection name does not
  # exist in the Connections Definition File.
  pass

class AdapterNotInstalled (Error):
  # Raised if a provider is requested for which
  # the python libraries are not installed.
  pass

class InvalidFormatError (Error):
  # Raised if the Connections Definition File is
  # in an unreadable format.
  pass


class GConnections:

  def __init__(self, location, loginHandler=None):

    self._loginHandler = loginHandler
    self._parser = ConfigParser()
    self._location = location

    GDebug.printMesg(1,'Conn File: "%s"' % location)

    if len(location):
      fileHandle = openResource(location)

      try:
        self._parser.readfp(fileHandle)
      except DuplicateSectionError:
        raise InvalidFormatError, \
           _('The connections file has duplicate source definitions.\n\nFile: %s') \
              % location
      except MissingSectionError:
        raise InvalidFormatError, \
           _('The connections file has no source definitions.\n\nFile: %s') \
              % location
      except:
        raise InvalidFormatError, \
           _('The connections file cannot be parsed.\n\nFile: %s') % location

    self._openConnections = {}

    self._primaries = {}
    self._aliases = {}
    self._definitions = {}


    # Read all the sections into a dict
    # and make a note of all alias names
    for section in self._parser.sections():
       self._primaries[section]={}
       for att in self._parser.options(section):
         if att == 'aliases':
           for alias in string.split(string.lower(self._parser.get(section,att))):
             self._aliases[alias] = section
         else:
           self._primaries[section][att] = self._parser.get(section, att)

    # Fill in any aliases with their parameters
    for alias in self._aliases.keys():
      section = self._aliases[alias]
      self._aliases[alias] = copy.copy(self._primaries[section])
      self._aliases[alias]['_alias_of'] = section


    self._definitions.update(self._aliases)
    self._definitions.update(self._primaries)



  def setLoginHandler(self, loginHandler):
    self._loginHandler = loginHandler


  def hasConnectionParameters(self, connection_name):
    return self._definitions.has_key(connection_name)


  def getConnectionParameter(self, connection_name, attribute, default=None):
    try:
      definition = self._definitions[connection_name]
      try:
        return definition[attribute]
      except:
        return default
    except KeyError:
      raise NotFoundError, \
         _('The connections file does not contain a definition \n       for "%s".\n\nFile: %s') \
            % (connection_name, self._location)


  #
  # Returns an dictionary of dictionaries describing all connections:
  #  {connection name: {att name: value}}
  #
  def getConnectionNames(self, includeAliases=1):
    if includeAliases:
      return self._definitions.keys()
    else:
      return self._primaries.keys()


  #
  # Returns an dictionary of dictionaries describing all connections:
  #  {connection name: {att name: value}}
  #
  def getAllConnectionParameters(self, includeAliases=1):
    if includeAliases:
      return copy.deepcopy(self._definitions)
    else:
      return copy.deepcopy(self._primaries)


  #
  # Returns a dictionary describing a connection:
  #  {att name: value}
  #
  def getConnectionParameters(self, connection_name):
    try:
      return copy.deepcopy(self._definitions[connection_name])
    except KeyError:
      raise NotFoundError, \
         _('The connections file does not contain a definition \n       for "%s".\n\nFile: %s') \
            % (connection_name, self._location)


  #
  # Add a connection entry (session specific; i.e., doesn't add
  # to the connections.conf file, but to the current instance's
  # list of available connections.
  #
  def addConnectionSpecification (self, name, parameters):
    self._definitions[string.lower(name)] = copy.copy(parameters)


  #
  # Return a database provider object
  #
  def getDataObject(self, connection_name, connection_type):
    # This will throw a GConnections.NotFoundError if an unknown
    # connection name is specified.  The calling method should
    # catch this exception and handle it properly (exit w/message)
    return _load_dbdriver(self.getConnectionParameter(connection_name, \
        'provider'), connection_type, self)


  #
  # Has a connection been initialized/established?
  #
  def isConnectionActive(self, connection):
    return self._openConnections.has_key(string.lower(connection))


  #
  # Get a data connection for a specified database
  #
  def requestConnection(self, dataObject, connection):
    connection_name = string.lower(connection)
    if self._openConnections.has_key(connection_name):

      # If a database connetion has already been established, use it
      dataObject.setDataConnection(self._openConnections[connection_name])
      GDebug.printMesg(5,'Reusing data connection to %s' % connection_name)

    else:
      
      # Get the parameters from the Connections Definition File
      loginData = self.getConnectionParameters(connection_name)
      netrcData = None
      try:
        # load the user's netrc file
        netrcData = netrc.netrc().authenticators("'gnue://%s/%s/%s'" %
                                                 ( loginData['host'],
                                                   loginData['provider'],
                                                   loginData['dbname'],
                                                   )
                                                 )
        GDebug.printMesg(5, 'Read the user\'s .netrc file')
      except IOError:
        pass

      if (netrcData):
        loginData['username'] = netrcData[0][1:-1]
        loginData['password'] = netrcData[2][1:-1]
        GDebug.printMesg(5, 'Found useful stuff for connection %s in the user\'s .netrc file' % connection_name)

      if (loginData.has_key('username')):
        loginData['_username'] = loginData['username']
        del loginData['username']

      if (loginData.has_key('password')):
        loginData['_password'] = loginData['password']
        del loginData['password']

      loginData.setdefault('_username')
      loginData.setdefault('_password')

      try:

        dataObject.connect(loginData)
        GDebug.printMesg(5, 'I had enough information to connect to %s without asking the user' % connection_name)
        # Ahora seteamos la coneccin para que pase a modo "Read committed"
        dataObject._dataConnection.cursor().execute("SET TRANSACTION ISOLATION LEVEL READ COMMITTED")

        # Save the newly opened connection for future datasources
        self._openConnections[connection_name] = dataObject.getDataConnection()

      except GDataObjects.LoginError:
        attempts = 4

        GDebug.printMesg(5,'Getting new data connection to %s' % connection_name)

        errortext = None
        while attempts:

          try:

            # Ask the UI to prompt for our login data
            loginData.update(self._loginHandler.getLogin(
              [connection_name,
               self.getConnectionParameter(connection_name,'comment',''),
               dataObject.getLoginFields()], errortext))

            # Ask the data object to connect to the database
            dataObject.connect(loginData)
            # Ahora seteamos la coneccin para que pase a modo "Read committed"
            self._dataConnection.cursor().execute("SET TRANSACTION ISOLATION LEVEL READ COMMITTED")

            # Save the newly opened connection for future datasources
            self._openConnections[connection_name] = dataObject.getDataConnection()

            # We're done!
            attempts = 0
            self._loginHandler.destroyLoginDialog()

          except GDataObjects.LoginError, error:
            # Oops, they must have entered an invalid user/password.
            # Those silly users.
            # user: Hey! Who are you calling silly?!!!
            attempts = attempts - 1
            errortext = string.replace("%s" % error,'\n','')

            if not attempts:
              # Four times is plenty...
              self._loginHandler.destroyLoginDialog()
              raise GDataObjects.LoginError, \
                    _("Unable to log in after 4 attempts.\n\nError: %s") % error

          except GLoginHandler.UserCanceledLogin:
            # Guess they changed their minds. Treat as a login error.
            self._loginHandler.destroyLoginDialog()
            raise GDataObjects.LoginError, _("User canceled the login request.")
        




#
# Load the correct DBdriver from gnue/common/dbdrivers/*/
#
def _load_dbdriver(driver, type, connectionManager):
  try:
    dbdriver = dyn_import("gnue.common.dbdrivers.%s" % (driver))
  except ImportError, err:
    GDebug.printMesg(1,'ImportError importing driver %s' % (driver))
    GDebug.printMesg(1,err)
    raise GDataObjects.ProviderNotSupportedError, \
       _("No database driver found for provider type '%s'") % driver

  try:
    dd = dbdriver.supportedDataObjects[type]()
    GDebug.printMesg (1,'Attaching to %s (%s)' % (driver,dd.__class__.__name__))
    return dd
  except KeyError:
    raise GDataObjects.ObjectTypeNotAvailableError, \
       _("DB Driver '%s' does not support source type '%s'") % (driver, type)


