#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GDataSource.py
#
# DESCRIPTION:
# Class that implements a provider-independent DataSource object
#
# NOTES:
#
# HISTORY:
#

import GDebug
import GDataObjects
import GObjects
import sys, string
import GConnections
import GTypecast
import GConditions


########################################################################
#
# Class that handles DataSources.  This is a subclass of GObj, which
# means this class can be created from XML markup and stored in an
# Object tree (e.g., a Forms tree).
#
########################################################################

class GDataSource(GObjects.GObj):

  def __init__(self, parent=None, type="GDataSource"):
    GObjects.GObj.__init__(self, parent, type)
    self.type = "object"
    self.database = None
    self._connections = None
    self._dataObject = None
    self._connectionComment = ""
    self._fieldReferences = {}
    self._unboundFieldReferences = {}
    self._defaultValues = {}

    self._inits =[self.primaryInit, self.secondaryInit, self.tertiaryInit]
    self._currentResultSet = None
    self._resultSetListeners = []
    self._toplevelParent = self._type # Needs to be set by subclass
                                      # so that _topObject gets set
    self._topObject = None

    #
    # trigger support
    #
    self._triggerGlobal = 1
    self._triggerFunctions = {'createResultSet':{'function':self.createResultSet,
                                                 },
                              'createRawResultSet':{'function':self.createRawResultSet,
                                                 },
                              'simpleQuery':{'function':self.triggerSimpleQuery,
                                             },
                              'getRecords':{'function':self.getRecords},
                              'getCurrentRecord':{'function':self.getCurrentRecord},
                              'getTimeStamp':{'function':self.getTimeStamp},
                              'getUnixTimeStamp':{'function':self.getUnixTimeStamp},
                              }

  def __getattr__(self, attr):
    if self._dataObject and attr[1] != '_' and hasattr(self._dataObject,attr):
      return getattr(self._dataObject,attr)
    else:
      raise AttributeError, attr

  def getCurrentResultSet(self):
    return self._currentResultSet
  
  def getRecords(self):
    lResultSet = self.getCurrentResultSet()
    if lResultSet._currentRecord == -1:
      return 0
    else:
      while lResultSet._loadNextRecord():
        pass
    return lResultSet.getCachedRecords()

  def getCurrentRecord(self):
    return self.getCurrentResultSet().getCurrentRecord()
  
  def getTimeStamp(self):
    return self._dataObject.triggerExtensions.getTimeStamp()

  def getUnixTimeStamp(self):
    return self._dataObject.triggerExtensions.getUnixTimeStamp()

  def triggerSimpleQuery(self,maskDict):
    queryDict = {}
    okToProcess = 1
    for key in maskDict.keys():
      queryDict[key]= str(maskDict[key])
      if not len(queryDict[key]):
        okToProcess = 0
        break

    conditions = GConditions.buildConditionFromDict(queryDict,GConditions.GClike)
    resultSet = self.createResultSet(conditions)
    recordCount = resultSet.getRecordCount()

    returnList = []
    for count in range(recordCount):
      record = resultSet.getRecord(count)
      resultDict = {}
      for key in record._fields.keys():
        resultDict[key]=record.getField(key) or ""
      returnList.append(resultDict)
    return returnList


  #
  # This method should be called after the object is created
  # but before any other methods are called
  #
  def setConnectionManager(self, connectionManager):
    self._connections = connectionManager

  def initialize(self):
    if not self.database:
      # We are a connectionless datasource (virtual?)
      # We have to bind to something, so bind to empty driver
      from gnue.common.dbdrivers._empty import DBdriver
      GDebug.printMesg (7, 'Using empty data driver')
      dataObject = DBdriver.supportedDataObjects['object']()

    elif self._connections:
      self.database = string.lower(self.database)
      # This will throw a GConnections.NotFoundError if an unknown
      # connection name is specified.  The calling method should
      # catch this exception and handle it properly (exit w/message)
      dataObject = \
         self._connections.getDataObject(self.database, self.type)
      GDebug.printMesg (7, "GDataSource.py bound to %s " % self._dataObject)

    self.name = string.lower(self.name)
    self._topObject._datasourceDictionary[self.name]=self

    dataObject._fieldReferences = self._fieldReferences
    dataObject._unboundFieldReferences = self._unboundFieldReferences
    dataObject._defaultValues = self._defaultValues
    dataObject._dataSource = self
    dataObject._outerJoinCondition = None
    dataObject._outerJoin = None
    for child in self._children:
      if isinstance(child, GConditions.GCondition):
        dataObject._staticCondition = child
        break
    for child in self._children:
      if isinstance(child, GConditions.GJoinCondition):
        dataObject._outerJoinCondition = child
        break

    # Copy all attributes from XML to the dataObject
    tagAttributes = getXMLelements()['datasource']['Attributes']
    for attribute in tagAttributes.keys():
      if self.__dict__.has_key(attribute):
        dataObject.__dict__[attribute] = self.__dict__[attribute]
      else:
        try:
          dataObject.__dict__[attribute] = tagAttributes[attribute]['Default']
        except KeyError:
          pass
    self._dataObject = dataObject

  def connect(self):
    if self.database != None:
      self._connections.\
           requestConnection(self._dataObject, self.database)

  def getDataObject(self):
    return self._dataObject

  def referenceField(self, field, defaultValue=None):
    GDebug.printMesg(7,'Field %s implicitly referenced' % field)
    self._fieldReferences[field] = ""

    if defaultValue != None:
      self._defaultValues[field] = defaultValue

  def referenceUnboundField(self, field, defaultValue=None):
    GDebug.printMesg(7,'Field %s implicitly referenced' % field)
    self._unboundFieldReferences[field] = 1

    if defaultValue != None:
      self._defaultValues[field] = defaultValue


  #
  # The following are a simple wrapper arround the datasource's dataobject
  # to hide the dataobject from the app programmer
  #
  def hasMaster(self):
    return self._dataObject != None and self._dataObject.hasMaster()

  def createRawResultSet(self, statement):
      resultSet= self._dataObject.createRawResultSet(statement)
      self.__setResultSet( resultSet )
      return resultSet

  def createResultSet(self, conditions={}, readOnly=0):
      resultSet= self._dataObject.createResultSet(conditions,readOnly)
      self.__setResultSet( resultSet )
      return resultSet

  def addDetailDataObject(self, dataObject, handler=None):
      self._dataObject.addDetailDataObject(dataObject, handler)

  def createEmptyResultSet(self, readOnly=0,masterRecordSet=None):
      resultSet = self._dataObject.createEmptyResultSet(readOnly, masterRecordSet=masterRecordSet)
      self.__setResultSet( resultSet )
      return resultSet

  #
  # Master/detail stuff
  #

  # Called by dbdrivers whenever this datasource's master has changed
  def masterResultSetChanged(self, masterResultSet, detailResultSet):
    self._masterResultSet = masterResultSet
    self.__setResultSet( detailResultSet )

  def __setResultSet(self, resultSet):
    self._currentResultSet = resultSet
    # Notify all the listeners (i.e., blocks) that the result set changed
    for listener in self._resultSetListeners:
      listener(resultSet)

  def registerResultSetListener(self, listener):
    self._resultSetListeners.append(listener)

  def primaryInit(self):
    self._topObject = self.findParentOfType(self._toplevelParent)
    GDebug.printMesg(10,"Setting %s to connect mgr %s" %(self.name,self._topObject._connections))
    self.setConnectionManager(self._topObject._connections)
    self.initialize()
    self.connect()
    # TODO: For now, adding to dtsrc to avoid a global namespace conflict (jcater)
    # TODO: I have not a clue to do with this at this time...originally from GFDataSource.py
    self.extensions = self._dataObject.triggerExtensions

  # TODO: Merged into GDataSource per the TODOs in reports and forms however
  # TODO: self._topObject._datasourceDictionary implies that the top object
  # TODO: always has a specifc structure.  This is a bad thing :(  Maybe GRootObj
  # TODO: should contain a getDatasourceDict()?
  #
  def secondaryInit(self):

    if hasattr(self, 'master') and self.master:

      self.master = string.lower(self.master)
      GDebug.printMesg(3,"Linking detail '%s' to master '%s'" \
                         % (self.name, self.master) )

      if self._topObject._datasourceDictionary.has_key(self.master):
        self._topObject._datasourceDictionary[self.master] \
            .getDataObject().addDetailDataObject(self.getDataObject(),
                                                 self)
      else:
        raise StandardError, \
           _("Detail source '%s' references non-existant master '%s'") \
                         % (self.name, self.master)

  def tertiaryInit(self):
    if hasattr(self, 'prequery'):
      if not self.hasMaster() and self.prequery:
        self.createResultSet()

  #
  # Hooks for record-level triggers
  #

  def _beforeCommitInsert(self, record):
    return 1

  def _beforeCommitUpdate(self, record):
    return 1

  def _beforeCommitDelete(self, record):
    return 1

  def _onModification(self, record):
    return 1

  def _onNewRecordQueried(self, record):
    return 1


######
#
#
#
######
class GSql(GObjects.GObj):
  def __init__(self, parent=None):
     GObjects.GObj(self, parent, type="GDSql")

######
#
# Used by client GParsers to automatically pull supported xml tags
#
######

#
# Return any XML elements associated with
# GDataSources.  Bases is a dictionary of tags
# whose values are update dictionaries.
# For example: bases={'datasource': {'BaseClass':myDataSource}}
# sets xmlElements['datasource']['BaseClass'] = myDataSource
#
def getXMLelements(updates={}):

  xmlElements = {
      'datasource': {
         'BaseClass': GDataSource,
         'Attributes': {
            'name':        {
               'Required': 1,
               'Unique':   1,
               'Typecast': GTypecast.name },
            'type':        {
               'Typecast': GTypecast.name,
               'Default':  "object" },
            'database':    {
               'Typecast': GTypecast.name },
            'table':       {
               'Typecast': GTypecast.name },
            'historytable':       {
               'Typecast': GTypecast.name },
            'posttype':       {
               'Typecast': GTypecast.name },
            'cache':       {
               'Typecast': GTypecast.whole,
               'Default':  5 },
            'prequery':    {
               'Typecast': GTypecast.boolean,
               'Default':  0 },
            'order_by':    {
               'Typecast': GTypecast.text },
            'master':      {
               'Typecast': GTypecast.name },
            'masterlink':  {
               'Typecast': GTypecast.text },
            'detaillink':  {
               'Typecast': GTypecast.text } },
         'ParentTags': None },
  }

  # Add conditional elements
  xmlElements.update(
      GConditions.getXMLelements(
          {'conditions':{'ParentTags':('datasource',) } } ))

  for alteration in updates.keys():
    xmlElements[alteration].update(updates[alteration])

  return xmlElements



