#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
#
# FILE:
# GObjects.py
#
# DESCRIPTION:
# Base class for GNUe objects which can be represented as XML
#
# NOTES:
#
import sys

try:
  from xml.sax import saxutils
except ImportError:
  print """
   This GNUe tool requires PyXML to be installed.
   Please download and install PyXML from:

      http://pyxml.sourceforge.net/

"""
  sys.exit()

import GDebug
import string
import types
from GParserHelpers import GContent
import GTypecast
from GTriggerCore import GTriggerCore

#
# Class GObj
#
# Base class for GNUe objects which can be represented as XML
#
class GObj(GTriggerCore):
  def __init__(self, parent=None, type='GObj'):
    GTriggerCore.__init__(self)
    self._type = type
    self._parent = parent       # The object that contains this object
    self._children = []         # The objects contained by this object
    self._attributes = {}
    self._inits = []            # functions called during phaseInit stage
    self._xmlnamespace = None
    if parent :
      parent.addChild(self)


  # This is a convenience function for applications
  # NOT using GParser to load an object tree.
  def buildObject(self, **params):
    self.__dict__.update(params)
    return self._buildObject()

    
  # This is a convenience function for applications
  # NOT using GParser to load an object tree.
  def buildAndInitObject(self, **params):
    self.phaseInit(self.buildObject(**params))



  #
  # phaseInit
  #
  def phaseInit(self, iterations=5, params={}):
    for phase in range(iterations):
      self._phaseInit(phase, params)

  def _phaseInit(self,phase,params):
    if (len(self._inits) > phase) and self._inits[phase]:
      GDebug.printMesg(6,"%s: Init Phase %s" % (self._type, phase+1))
      if self._type=='GFForm':
        self._inits[phase](params)
      else:
        self._inits[phase]()

    for child in self._children:
      if isinstance(child, GObj):
         child._phaseInit(phase,params)

  # This function is called after the parsers have completely
  # constructed. All children should be in place and
  # attributes and content should be set at this point.
  # Return the number of phaseInit iterations your object will
  # need.
  #
  # NOTE: Do not initialize datasources, etc at this point.
  #       This is only so content can be set, etc, after
  #       loading from XML.
  #
  def _buildObject(self):
    return len(self._inits)

  def getChildrenAsContent(self):
    content = ""
    for child in self._children:
      if isinstance(child, GContent):
        content += child._content
    return content

  def showTree(self, indent=0):
    print ' ' * indent + `self._type`
    for child in self._children:
       child.showTree(indent + 2)

  #
  # addChild
  #
  # Adds an object to an instances list of children
  #
  def addChild(self, child):
    self._children.append(child)

  def toXML(self):
    xml_outer = string.lower(self._type[2:])
    r = "<" + xml_outer + ">\n  <options>\n"
    for k in self.__dict__.keys():
      # skip keys beginning with _
      if k[0] == "_":
        continue
      val = self.__dict__[k]
      if isinstance(val, _types.StringType):
        str = val
      elif isinstance(val, _types.IntType) or isinstance(val, _types.FloatType):
        str = repr(val)
      else:
        continue
      r += "    <" + k + ">" + str + "</" + k + ">\n"
    r += "  </options>\n</" + xml_outer + ">\n"
    return r

  #
  # dumpXML
  #
  # Dumps an XML representation of the object
  #

  def getXmlTag(self):
    return string.lower(string.replace(self._type[2:],'_','-'))

  def dumpXML(self, lookupDict, treeDump=None, gap="  ", xmlnamespaces={}):
    xmlns = ""
    xmlnsdef = ""
    try:
      if self._xmlnamespace:
        try:
          xmlns = xmlnamespaces[self._xmlnamespace] + ":"
        except KeyError:
          i = 0
          xmlns = "out"
          while xmlns in xmlnamespaces.values():
            i += 1
            xmlns = "out%s" % i
          xmlnamespaces[self._xmlnamespace] = xmlns
          xmlnsdef = ' xmlns:%s="%s"' % (xmlns, self._xmlnamespace)
          xmlns += ":"
    except AttributeError:
      try:
        if self._xmlchildnamespace:
          if not xmlnamespaces.hasattr(self._xmlchildnamespace):
            i = 0
            ns = "out"
            while ns in xmlnamespaces.values():
              i += 1
              ns = "out%s" % i
            xmlnamespaces[self._xmlnamespace] = ns
            xmlnsdef = ' xmlns:%s="%s"' % (ns, self._xmlchildnamespace)
      except AttributeError:
        pass

    xmlEntity = self.getXmlTag()
    xmlString = "%s<%s%s%s" % (gap[:-2],xmlns, xmlEntity, xmlnsdef)

    indent = len(xmlString)
    pos = indent
    attrs = self.__dict__.keys()
    attrs.sort()
    for attribute in attrs:
      # skip keys beginning with _
      if attribute[0] == "_":
        continue
      val = self.__dict__[attribute]
      if (xmlns and attribute in self._listedAttributes) or \
         (not xmlns and lookupDict[xmlEntity].has_key('Attributes') and \
         lookupDict[xmlEntity]['Attributes'].has_key(attribute)):
        if val != None and \
           (xmlns or (not lookupDict[xmlEntity]['Attributes'][attribute].has_key('Default') or \
            (lookupDict[xmlEntity]['Attributes'][attribute]['Default']) != (val))):
          try:
            typecast = lookupDict[xmlEntity]['Attributes'][attribute]['Typecast']
          except:
            typecast = GTypecast.text
          if typecast == GTypecast.boolean \
             and val == 1:
            addl = ' %s=""' % (attribute)
          elif typecast == GTypecast.names:
            addl = ' %s="%s"' % \
                (attribute, string.join(val,','))
          else:
            addl = ' %s="%s"' % (attribute, saxutils.escape('%s' % val))
          if len(addl) + pos > 78:
            xmlString += "\n" + " " * indent + addl
            pos = indent
          else:
            xmlString += addl
            pos += len(addl)

    if len(self._children):
      hasContent = 0
      for child in self._children:
        hasContent = hasContent or isinstance(child,GContent)
      if hasContent:
        xmlString += ">"
      else:
        xmlString += ">\n"

      if treeDump:
        for child in self._children:
          xmlString += child.dumpXML(lookupDict, 1,gap+"  ",
              xmlnamespaces=xmlnamespaces)

      if hasContent:
        xmlString += "</%s%s>\n" % (xmlns, xmlEntity)
      else:
        xmlString += "%s</%s%s>\n" % (gap[:-2], xmlns,xmlEntity)
    else:
      xmlString += "/>\n"
    return xmlString


  def walk(self, function, *args, **parms):
    function(self, *args, **parms)
    for child in self._children:
      if isinstance(child, GObj):
        child.walk(function, *args, **parms)

  #
  # findParentOfType
  #
  # Moves upward though the parents of an object till
  # it finds the parent of the specified type
  #
  def findParentOfType(self,type, includeSelf=1):

    if includeSelf:
      parentObject = self
    else:
      parentObject = self._parent
#
# Replacing this code with code below decreased startup
# function call time by approx 33%
#
#      if parentObject is None:
#        return None
#
#    while (parentObject._parent != None and parentObject._type != type):
#      parentObject = parentObject._parent
#
#    if parentObject._type == type:
#      return parentObject
#    else:
#      return None

    while 1:
      if parentObject == None:
        return None
      elif parentObject._type == type:
        return parentObject

      parentObject = parentObject._parent



  #
  # getDescription
  #
  # Return a useful description of this object
  # Used by designer clients
  #
  def getDescription(self):
    if hasattr(self,'_description'):
      return self._description
    elif hasattr(self,'name'):
      return self.name + " (%s)" % self._type[2:]
    else:
      return self._type[2:] + " (%s)" % self._type[2:]

