#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GParser.py
#
# DESCRIPTION:
# Class that contains a SAX2-based XML processor for GNUe
#
# NOTES:
#

import sys

try:
  from xml.sax import saxutils
  import xml.sax
except ImportError:
  print """
   This GNUe tool requires pythons XML module be installed.
   Typically this is the case, however some GNU/Linux distro's
   like Debian distribute this as a seperate package

   To install this package...
     On Debian: apt-get install python-xml

"""


import string
import GDebug
import GTypecast
from GParserHelpers import GContent


#######################################################
#
# Error classed raised for markup errors
#
class MarkupError(StandardError):
  pass


#######################################################
#
# loadXMLObject
#
# This method loads an object from an XML file and
# returns that object.  If initialize is 1 (default),
# then the object is initialized and ready to go.
# Setting initialize to 0 is useful for a design
# environment where the object is not actually
# being used, but simply loaded as a document.
#
# "attributes" is a dictionary containing extra
# attributes that should be attached to this object.
#
#  e.g., if attributes={myproperty:[0,1,2]}, then
#    before the object is initialized or returned,
#    object.myproperty == [0,1,2].
#
#######################################################

def loadXMLObject(stream, handler, rootType, xmlFileType,
  initialize=1, attributes={}, initParameters={}):
  # Create a parser
  parser = xml.sax.make_parser()

  # Set up some namespace-related stuff for the parsers
  parser.setFeature(xml.sax.handler.feature_namespaces, 1)

  # Create a stack for the parsing routine
  object = None

  # Create the handler
  dh = handler()

  # Tell the parser to use our handler
  parser.setContentHandler(dh)
  parser.parse(stream)

  object = dh.getRoot()

  if not object:
    raise MarkupError, _("Error loading %s: empty definition file") % (xmlFileType)
  elif object._type != rootType:
    raise MarkupError, _("Error loading %s: not a valid %s definition (expected: %s, got: %s)") % (xmlFileType,
         xmlFileType, rootType, object._type)

  # Set the object's attributes
  object.__dict__.update(attributes)

  if initialize:
    GDebug.printMesg(10,"Initializing the object tree starting at %s" %(object))
    object.phaseInit(dh._phaseInitCount, initParameters)

  return object



#######################################################
#
# char
#
# This is for typecasting strings
#
# NOTE: This is a redefinition of GTypecast.text.
#   This redefinition will be removed as soon as all
#   references to it are changes.  You should be using
#   GTypecast from now on.
#
#######################################################
char  = GTypecast.text


#######################################################
#
# bool
#
# This is for typecasting booleans
#
# NOTE: This is a redefinition of GTypecast.boolean.
#   This redefinition will be removed as soon as all
#   references to it are changes.  You should be using
#   GTypecast from now on.
#
#######################################################
bool = GTypecast.boolean


#######################################################
#
# normalise_whitespace
#
# Remove redundant whitespace from a string ( from xml-howto )
#
#######################################################
def normalise_whitespace(text):
  return string.join( string.split(text), ' ')


def default(attrs, key, default):
  try:
    return attrs[key]
  except KeyError:
    return default


#######################################################
#
# xmlHandler
#
# This class is called by the XML parser to
# process the xml file.
#
#######################################################
class xmlHandler(xml.sax.ContentHandler):
  def __init__(self):

    self.xmlElements = {}
    self.xmlMasqueradeNamespaceElements = None

    self.xmlStack = []
    self.nameStack = []
    self.bootstrapflag = 0
    self.uniqueIDs = {}
    self.root = None
    self._phaseInitCount = 0

  #
  # Called by client code to get the "root" node
  #
  def getRoot(self):
    return self.root

  #
  # Called by the internal SAX parser whenever
  # a starting XML element/tag is encountered.
  #
  def startElementNS(self, qtag, qname, saxattrs):
    ns, name = qtag
    attrs = {}
    loadedxmlattrs = {}

    if not ns:
      #
      # No namespace qualifier
      #
      GDebug.printMesg(50, "<%s>" % name)

      try:
        baseAttrs = self.xmlElements[name].get('Attributes',{}) # default(self.xmlElements[name],'Attributes',{})
      except KeyError:
        raise MarkupError, _('Error processing <%s> tag [I do not know what a <%s> tag does]') % (name, name)


      for qattr in saxattrs.keys():
        attrns, attr = qattr


        # Typecasting, anyone?  If attribute should be int, make it an int
        try:
          attrs[attr] = baseAttrs[attr].get('Typecast',char)(saxattrs[qattr]) # default(baseAttrs[attr],'Typecast',char)(saxattrs[qattr])
          loadedxmlattrs[attr] = attrs[attr]
        except KeyError:
          raise MarkupError, _('Error processing <%s> tag [I do not recognize the "%s" attribute') % (name, attr)
        except:
          raise MarkupError, _('Error processing <%s> tag [invalid type for "%s" attribute; value is "%s"]') % (name, attr, saxattrs[qattr])

        # If this attribute must be unique, check for duplicates
        if baseAttrs[attr].get('Unique',0): # default (baseAttrs[attr],'Unique',0):
          if self.uniqueIDs.has_key('%s' % (saxattrs[qattr])):
            raise MarkupError, _('Error processing <%s> tag ["%s" attribute should be unique; duplicate value is "%s"]') % (name, attr, saxattrs[qattr])

      for attr in baseAttrs.keys():
        if not attrs.has_key(attr):

          # Pull default values for missing attributes
          if baseAttrs[attr].has_key ('Default'):
            attrs[attr] = baseAttrs[attr].get('Typecast', char) (baseAttrs[attr]['Default'])# default(baseAttrs[attr],'Typecast', char) (baseAttrs[attr]['Default'])

          # Check for missing required attributes
          elif baseAttrs[attr].get('Required', 0): #default(baseAttrs[attr], 'Required', 0):
            raise MarkupError, _('Error processing <%s> tag [required attribute "%s" not present]') % (name, attr)


      if self.bootstrapflag:
        if self.xmlStack[0] != None:
          object = self.xmlElements[name]['BaseClass'](self.xmlStack[0])
      else:
        object = self.xmlElements[name]['BaseClass']()
        self.root = object
        self.bootstrapflag = 1

      object._xmltag = name

    elif self.xmlMasqueradeNamespaceElements:
      #
      # namespace qualifier and we are masquerading
      #

      GDebug.printMesg(50, "<%s:%s>" % (ns,name))

      for qattr in saxattrs.keys():
        attrns, attr = qattr

        attrs[attr] = saxattrs[qattr]
        loadedxmlattrs[attr] = saxattrs[qattr]

      object = self.xmlMasqueradeNamespaceElements(self.xmlStack[0])
      object._xmltag = name
      object._xmlnamespace = ns
      object._listedAttributes = loadedxmlattrs.keys()

    else:
      #
      # namespace qualifier and we are not masquerading
      #
      print _("WARNING: Markup includes namespaces, but the current tool does not include namespace support!")
      sys.exit()


    # Save the attributes loaded from XML file
    # (i.e., attributes that were not defaulted)
    object._loadedxmlattrs = loadedxmlattrs

    # Set the attributes
    object.__dict__.update(attrs)

    self.xmlStack.insert(0, object)
    self.nameStack.insert(0, name)


  #
  # Called by the internal SAX parser whenever
  # text (not part of a tag) is encountered.
  #
  def characters(self, text):

    if self.xmlStack[0] != None:

      # Masqueraging namespace elements, then keep content
      xmlns = self.xmlMasqueradeNamespaceElements and \
          isinstance(self.xmlStack[0],self.xmlMasqueradeNamespaceElements)

      # Should we keep the text?
      if xmlns or self.xmlElements[self.nameStack[0]].get('MixedContent',0):

        if xmlns or self.xmlElements[self.nameStack[0]].get('KeepWhitespace',0):
          GContent(self.xmlStack[0], text)
        else:
          # Normalize
          if len(string.replace(string.replace(string.replace(text,' ',''),'\n',''),'\t','')):
            text = normalise_whitespace (text)
          else:
            text = ""
          if len(text):
            GContent(self.xmlStack[0], text)


  #
  # Called by the internal SAX parser whenever
  # an ending XML tag/element is encountered.
  #
  def endElementNS(self, qtag, qname):
    ns, name = qtag
    self.nameStack.pop(0)
    child = self.xmlStack.pop(0)
    inits = child._buildObject()
    self._phaseInitCount = (inits != None and inits > self._phaseInitCount \
                            and inits or self._phaseInitCount)
    GDebug.printMesg(50, "</%s>" % name)


