#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it 
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either 
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied 
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public 
# License along with program; see the file COPYING. If not, 
# write to the Free Software Foundation, Inc., 59 Temple Place 
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
#
# FILE:
# GTrigger.py
#
# DESCRIPTION:
# Provides the basic classes needed by the generic trigger system
#
# NOTES:
#
import sys
import types
import string
from gnue.common.GObjects import GObj
from gnue.common import GDebug

from gnue.common import GTypecast
from xml.sax import saxutils
from gnue.common.GParser import GContent

class TriggerError:
  def __init__(self, msg):
    self.msg = msg


class TriggerAbort:
  pass
class TriggerStop:
  pass
class TriggerSuccess:
  pass


        

#######################################################################
#
# Trigger namespace classes
#
# Classes used in implementing the trigger namespace
#
#
# GTriggerNamespace
#
# Manager class for a namespace.  Passed a GObj based
# tree at __init__ which it uses to construct a new
# GTriggerNSObject based tree
#
class GTriggerNamespace(GObj):
  def __init__(self,objectTree = None):
    self._globalNamespace = {'True' : 1,
                             'False': 0,
                             }
    if objectTree:
      self._globalNamespace['form'] = self.constructTriggerObject(objectTree)
    else:
      GDebug.printMesg(0,'GTriggerNamespace was passed an empty object tree')

  #
  # constructTriggerObject
  #
  # Travels down thru a GObj based tree and builds a set
  # of GTriggerNSObjects that will implement the namespace 
  # inside triggers.
  # 
  def constructTriggerObject(self, gobjObject, triggerParent=None):
    triggerObject = None

    # Some items in a GObj tree may not be GObj based (GContent for instance)
    
    if isinstance(gobjObject,GObj) and hasattr(gobjObject,'name'): 
      triggerObject = GTriggerNSObject(triggerParent)
      triggerObject._object = gobjObject

      # Add this triggerObject to global namespace if the GObj requests it
      if gobjObject._triggerGlobal:
        self._globalNamespace[gobjObject.name] = triggerObject

      # setup get and set functions when they exist in the GObj
      triggerObject._triggerSet = gobjObject._triggerSet
      triggerObject._triggerGet = gobjObject._triggerGet
      
      # Add any trigger methods defined by GObj 
      if len(gobjObject._triggerFunctions):
        for item in gobjObject._triggerFunctions.keys():
          
          if type(gobjObject._triggerFunctions[item]['function']) == types.MethodType:
            object = GTriggerNSFunction(item,gobjObject._triggerFunctions[item]['function'])                        
            triggerObject.__dict__[item] = object
            # Add this function to global namespace if the GObj requests it
            if gobjObject._triggerFunctions[item].has_key('global') and \
               gobjObject._triggerFunctions[item]['global'] != 0:
              self._globalNamespace[item] = object

          else:
            GDebug.printMesg(0,'Only functions are supported in an objects _triggerFunctions %s %s' % (gobjObject,item))
            
            sys.exit()

      # Load the defined __properties__ into this object's
      # GTriggerNSObjectProperties instance
      if len(gobjObject._triggerProperties):
        for item in gobjObject._triggerProperties.keys():
          if gobjObject._triggerProperties[item].has_key('set'):
            setFunc = gobjObject._triggerProperties[item]['set']
          else:
            setFunc = None
          triggerObject._triggerProperties.addProperty(item,gobjObject._triggerProperties[item]['get'], setFunc)

      # Process the children of this Gobj 
      if len(gobjObject._children):
        for child in gobjObject._children:
          object = self.constructTriggerObject(child, triggerObject)
          
          # Add this objects children to it's namespace by their name
          if object:
            triggerObject.__dict__[child.name] = object

      #
      # populate the GObj's _localTriggerNamespace
      localNamespace = {'self':triggerObject}
      localNamespace.update(triggerObject.__dict__)
      gobjObject._localTriggerNamespace = localNamespace

    return triggerObject

#
# GTriggerNSObject
#
# Inherrits GObj to gain it's parent/child system
#
class GTriggerNSObject(GObj):
  def __init__(self, parent):
    GObj.__init__(self,parent)
    self._triggerProperties = GTriggerNSObjectProperties()
    self._triggerSet = None
    self._triggerGet = None
    self._object = None
    
  #
  # showTree
  #
  # Handy function to dump a rough look at the namespace
  # doesn't show things nested properly though
  def showTree(self, indent=0):
    print '  ' * indent + `self._type` + `self`
    for item in self.__dict__.keys():
      if item[:1] != '_':
        print '  ' * indent + ' :' + item
    for child in self._children:
      child.showTree(indent + 2)

  #
  # __setattr__
  #
  # This is called when trying to write something inside a trigger object
  # namespace.  It checks to see if the var name is already linked to a
  # GTriggerNSObject based object and calls that objects _triggerSet if it
  # exists.
  #
  # Example: form.block1.entry1 = "foo"
  #
  # The __setattr__ will execute at the block1 and call the functions that
  # are part of the entry1 object to set it's value
  #
  def __setattr__(self, name, value):
    if self.__dict__.has_key(name) and \
       isinstance(self.__dict__[name], GTriggerNSObject):
     if isinstance(self.__dict__[name], GTriggerNSObject):
      if self.__dict__[name]._triggerSet:
        self.__dict__[name]._triggerSet(value)
      else:
        GDebug.printMesg(0,'Trigger attempting to reset a form object')
    else:
      self.__dict__[name] = value

  #
  # __getattr__
  #
  # Only needed to return the GTriggerNSObjectProperties
  # object 
  #
  def __getattr__(self,name):
    if name == '__properties__':
      return self._triggerProperties
    else:
      print "AttributeError: %s" % (name)
      # print self.__dict__
      raise AttributeError
          
  #
  # __str__
  #
  # This executes at a different level than the __setattr__
  # While __setattr__ is executed in the parent object to protect
  # it's namespace object links, this routine is called by the
  # object referenced
  #
  # Example: foo = form.block1.entry1
  #
  # This __str__ would execute against the entry1 object
  #
  def __str__(self):
    if self._triggerGet:
      return str(self._triggerGet())
    else:
      return ""

  #
  # __cmp__
  #
  # Forces the system to compare the string values of
  # GTriggerNSObject objects and not their instances
  #
  def __cmp__(self,other):
    selfvalue = "%s" % self
    othervalue = "%s" % other
    
    if selfvalue == othervalue:
      return 0
    elif selfvalue < othervalue:
      return -1
    else:
      return 1


  #
  # __nonzero__
  #
  # Needed to make __len__ function below play nice
  #
  def __nonzero__(self):
    return 1

  #
  # __len__
  #
  # Implements len() support
  #
  def __len__(self):
    string =  "%s" % self
    return len(string)

  #
  # __getitem__
  #
  # implements support for sub selections of strings
  #
  # example: block1.fieldname[0:4]
  #
  def __getitem__(self, key) :
    string = self.__str__()
    return string[key.start:key.stop]

#
# GTriggerNSFunction
#
# Accessor class for functions that are made available in the trigger
# namespace
#     
class GTriggerNSFunction:
  def __init__(self, name, functionLink):
    self._name = name
    self._objectFunction = functionLink

  def __call__(self, *args):
    return self._objectFunction(*args)

#
# GTriggerNSObjectProperties
#
# Accessor class for properties that are made available in this an
# object's __properties__ namespace
#
class GTriggerNSObjectProperties:
  def __init__(self):
    self._properties = {}

  def addProperty(self,name, getFunc, setFunc):
    self._properties[name] = {'get':getFunc,
                              'set':setFunc,
                              }

  def __setattr__(self, name, value):
    # Hack to ensure that self._properties exists
    if not self.__dict__.has_key('_properties'):
      self.__dict__['_properties'] = {}
      
    if self._properties.has_key(name):
      # If none the it's readonly
      if self._properties[name]['set']: 
        self._properties[name]['set'](value)
      else:
        GDebug.printMesg(0,'Attempt to set readonly property :%s' %(name))
    else:
      self.__dict__[name] = value

  def __getattr__(self,name):
    if self.__dict__['_properties'].has_key(name):
      return self._properties[name]['get']()
    else:
      raise AttributeError






#######################################################################
#
# Trigger instance classes
#
# Classes in here are used to represent the triggers defined in
# <trigger> tags in an xml file
#
#
# Note: this is a partial graft of the old GFTrigger system
#       only here so that the new namespace code could be
#       put to use right away

#
# A list of all valid triggers, and their "pretty" names
#
VALIDTRIGGERS = { 'PRE-FOCUSOUT':   'Pre-FocusOut',
                  'POST-FOCUSOUT':  'Post-FocusOut',
                  'PRE-FOCUSIN':    'Pre-FocusIn',
                  'POST-FOCUSIN':   'Post-FocusIn',
                  'PRE-COMMIT':     'Pre-Commit',
                  'POST-COMMIT':    'Post-Commit',
                  'POST-QUERY':     'Post-Query',
                  'ON-SWITCH':      'On-Switch',
                  'PRE-CHANGE':     'Pre-Change',
                  'POST-CHANGE':    'Post-Change',
                  'ON-NEWRECORD':   'On-NewRecord' }




#
# GTrigger
#
# Class used to implement triggers
#
class GTrigger(GObj):
  def __init__(self, parent=None, type=None, name=None, src=None, text=None, language='python'):
    GObj.__init__(self, parent, 'GTrigger')

    self._text=''
    self._triggerns={}
    self._inits   = [None,self.initialize]

    self.language=language
    self.src = src
    self.type = type and string.upper(type) or None
    self.name = name
    if text != None:
      GContent(self, text)
    if self.type != None:
      self._buildObject()

  #
  # Must be at least a phase 2 init
  #
  # TODO: this function merges the local namespace
  # TODO: not sure if this is doing the right thing
  # TODO: with regard to NAMED triggers.  It should
  # TODO: merge the local namespace of the object that
  # TODO: fired the trigger.
  def initialize(self):
    self._form = self.findParentOfType('GFForm')
    self._triggerns.update( self._form._triggerns )
    self.__call__ = self.dummyFunction

    if self.type != "NAMED":
      if self._parent:
        self._parent.addTrigger( self.type, self )
        self._triggerns.update(self._parent._localTriggerNamespace)
    else:
      form = self.findParentOfType('GFForm')
      form._triggerDictionary[self.name] = self
      self._triggerns.update(form._localTriggerNamespace)

    if self.src == None: 
      self.setFunction( self.getChildrenAsContent(), self.language )
    else:
      self.setFunctionFrom(self._form._triggerDictionary[self.src])


  def setFunctionFrom(self, object):
    self.__call__=object.__call__
  
  def setFunction(self, text, language):
    self._text = text
    self.language = language
    if self.language != 'python':
      GDebug.printMesg(0, "Language %s not implemented" % self.language)
      sys.exit()


    # TODO: this isn't working?????
    # Verify that the trigger contains no tabs
    if string.find('\t', self._text) >= 0:
      GDebug.printMesg(0, "The trigger named %s contains a tab character which is not allowed at pos %s"
                       % ( self.name, string.find('\t', self._text) ))
      sys.exit()
      
    # Remove whitespace from last line
    self._text = string.rstrip(self._text)

    # First compile the trigger:
    #   Compilation at this stage has the useful side effect that
    #   syntax errors are spotted during XML parsing rather than
    #   during execution.

    # Get the indentation level of the first line of code so 
    # we can indent our imports to the same level
    indentLevel = 0
    for line in string.split(string.replace(self._text,'\r',''),'\n'):
      if len(string.strip(line)) and string.lstrip(line)[0] != '#':
        indentLevel = len(line) - len(string.lstrip(line))
        break

    revisedTriggerCode = \
      "from gnue.common.GTrigger import TriggerError\n"
    for line in string.split(string.replace(self._text,'\r',''),'\n'):
      revisedTriggerCode = revisedTriggerCode + line[indentLevel:] + '\n'
    revisedTriggerCode = revisedTriggerCode + \
      "pass\n"

    try:
      self._code = compile(revisedTriggerCode, '<trigger '+str(self.name)+'>', 'exec')
    except SyntaxError, err:
      for line in revisedTriggerCode.split ('\n'):
        GDebug.printMesg (0, ">%s<" % line)
      GDebug.printMesg(0, "[%s]" % self.name)
      GDebug.printMesg(0, "Syntax error in %s:%s [%s] of trigger in element ?? [%s] >%s<" \
           % ( str(err.lineno), str(err.offset), err.text, self.name, revisedTriggerCode.split ('\n')[err.lineno] ))
      sys.exit()
    except:
      # GDebug.printMesg(0, "[%s]" % self.name)
      print "[%s]" % self.name
      raise

    def thisTrigger(myself, code = self._code,
                    triggerns = self._triggerns):
#      triggerns['self'] = myself
#      triggerns['runform'] = myself.findParentOfType('GFForm')._app.getManager().runFormFromTrigger
#      triggerns.update(
      try:
        locals = {}
        exec code in triggerns, locals
      except TriggerError:
        GDebug.printMesg(0, "[%s]" % myself.name)
        raise
      except:
        # May be better to deal with this in GFTriggerAware
        GDebug.printMesg(0, "[%s]" % myself.name)
        raise
        import sys
        GDebug.printMesg(0, "%s in trigger code, value: %s" % (sys.exc_type, sys.exc_value))

    self.__call__ = thisTrigger

  def dummyFunction(self, myself):
    GDebug.printMesg(0, "Trigger not implemented")

  #
  # getDescription
  #
  # Return a useful description of this object
  # for use by designer
  #
  def getDescription(self):
    if self.type == 'NAMED': 
      return self.name
    else: 
      return VALIDTRIGGERS[string.upper(self.type)]

  #
  # dumpXML
  #
  # Dumps an XML representation of the object
  # used in saving
  #
  def dumpXML(self, lookupDict, treeDump=None, gap=None,xmlnamespaces={}):
    escape = not int(gConfig('StoreTriggersAsCDATA'))
    xmlEntity = "trigger"
    xmlString = "%s<%s" % (gap[:-2],xmlEntity)

    indent = len(xmlString) 
    pos = indent
    for attribute in self.__dict__.keys():
      
      # variables beginning with _ are never saved out to file
      # they are internal to the program
      if attribute[0] == "_":
        continue
      
      val = self.__dict__[attribute]
      if lookupDict[xmlEntity].has_key('Attributes') and \
         lookupDict[xmlEntity]['Attributes'].has_key(attribute): 
        if val != None and \
           (not lookupDict[xmlEntity]['Attributes'][attribute].has_key('Default') or \
            (lookupDict[xmlEntity]['Attributes'][attribute]['Default']) != (val)): 
          typecast = lookupDict[xmlEntity]['Attributes'][attribute]['Typecast']
          if typecast == GTypecast.boolean \
             and val == 1:
            addl = ' %s=""' % (attribute)
          elif typecast == GTypecast.names:
            addl = ' %s="%s"' % \
                (attribute, string.join(val,','))
          else: 
            addl = ' %s="%s"' % (attribute, saxutils.escape('%s' % val))
          if len(addl) + pos > 78: 
            xmlString += "\n" + " " * indent + addl
            pos = indent
          else: 
            xmlString = xmlString + addl
            pos += len(addl)
        
    if len(self._children): 
      hasContent = 0 
      for child in self._children: 
        hasContent = hasContent or isinstance(child,GContent)
      if hasContent: 
        xmlString += ">"
      else: 
        xmlString += ">\n"
       
      if treeDump:
        if hasContent and not escape: 
          xmlString += "<![CDATA["
        for child in self._children:
          xmlString += child.dumpXML(lookupDict, 1,gap+"  ",escape=escape)
        if hasContent and not escape: 
          xmlString += "]]!>"

      if hasContent: 
        xmlString += "</%s>\n" % (xmlEntity)
      else: 
        xmlString += "%s</%s>\n" % (gap[:-2], xmlEntity)
    else: 
      xmlString += "/>\n"      
    return xmlString




          
#######################################################################
#
# Trigger processor classes
#
# GTriggerExtension
#
# Objects that inherit this class will be
# capable of processing triggers
#
class GTriggerExtension:
  def __init__(self):
    self._trigger = {}
    self._validTriggers = {} # TODO : This probably needs moved into the apps trigger manager.
                             # TODO : I'll think about this more after sleep.

  # associateTrigger
  #
  # Associates a trigger with the object.  More than one trigger of a specific type
  # can be associated with an object
  #
  def associateTrigger(self, key, function):
    key = string.upper(key)
    if key in self._validTriggers:
      if not self._trigger.has_key(key):
        self._trigger[string.upper(key)] = []
      self._trigger[string.upper(key)].append(function)
    else:
      print _("Invalid trigger "),key

  # processTrigger
  # 
  # "fires" the trigger
  def processTrigger(self, key):
    key = string.upper(key)
    if key in self._validTriggers:
      if self._trigger.has_key(key):
        for function in self._trigger[key]:
          function(self)
      else:
        GDebug.printMesg(10, "No triggers to fire")
    else:
      print _("Invalid trigger "),key






