#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001 Free Software Foundation
#
# FILE:
# xmlrpc/ClientAdapter.py
#
# DESCRIPTION:
# Set of classes that implement the XML-RPC client adapter for GNUe Comm.
#
# NOTES:
# Requires xmlrpclib from http://www.pythonware.com/products/xmlrpc/
# or Python 2.2
#
# Client Parameters:
#
#    url         The complete URI address of the XML-RPC service.
#                e.g, url="https://www.domain.com:9876/service.php"
#    ..or..
#    transport   The service transport (either "http" or "https")
#    host        The hostname or IP address of the service
#    port        The port that the service is located on
#



#
# We provide a client driver...
#
CLIENT = 1      # ClientAdapter


from gnue.common.commdrivers import GCommBase
from gnue.common import GComm,GConfig

import string, sys

#
# i18n support
#
import gettext

try:
  gettext.install('gnue',GConfig.getInstalledBase('install_prefix')+'/translations')
except IOError:
  sys.stderr.write('Unable to load locale information falling back to default English language translations.\n')    
  import __builtin__
  __builtin__.__dict__['_'] = lambda string:string
  
try:

  # xmlrpclib could be in a package, so try both
  try:
    import xmlrpclib
  except ImportError:
    from xmlrpclib import xmlrpclib
  
except ImportError:
  raise GComm.AdapterInitializationError, \
        _("\nUnable to load xmlrpclib.  To use the XML-RPC interface, \n") \
      + _("please install xmlrpc from:\n") \
      + "    http://www.pythonware.com/products/xmlrpc/"


##############################################################################
#
# ClientAdapter
#
class ClientAdapter(GCommBase.Client):

  def __init__(self, params):


    try:

      try:
        url = params['url']
        if not len(url):
          raise KeyError
      except KeyError:
        try:
          transport = params['transport']
        except KeyError:
          transport = 'https'

        url = "%s://%s:%s/" % \
            ( transport, params['host'], params['port'] )



        # add an list to store references to proxys for dyn. objects
        self._objectProxys=[]

        self._server = xmlrpclib.Server(url)


    except KeyError:
      raise GComm.AdapterConfigurationError,  \
            _("To use XML-RPC, you must either specify a 'url' or\n") +\
            _("a 'host' and 'port' setting.")
    except:
      raise GComm.AdapterInitializationError, \
         _("Unable to initialize the XML-RPC interface at %s") % url


  def request(self, service, params={}):
    self._baseproxy = _ProxyObject(self, service, \
                                   servicer=self._server.__getattr__(service))   
    return self._baseproxy
  
  def close(self):    
    # do a cleanup of the proxy objects and the object stubs
    # on the XMLRPC server side

    # 1. close the proxy objects for services (static, client only)
    self._baseproxy=None    
    
    # 2. close the proxy objects for objects (dynamic, client+server)
    for i in self._objectProxys:
      i._close()      
    self._objectProxys=[] 
    
    # close the client it self
    self._client=None    

  # just in case the user forgot to delete some objects
  def __del__(self):
    if self._client!=None:
      self.close()
    GCommBase.Client.__del__()

  def setTimeout(self,timeout):
    self._timeout=timeout


  def runMethod(self, method, *args, **params):
    # print "Calling method: %s with attr %s " % method,string.join(args,",")

    to_call=getattr(self._server,method);
        
    result=to_call.__call__(*args, **params)
    # print "the result is",result,"of type",type(result)
    
    # check, if an object handle is send    
    # TODO: make a better check
    if type(result)==type("String") and len(result)==40 :
      # create an proxy for this handle      
      result = _ProxyObject(self, "[%s]" % result)

      # store proxy in list
      self._objectProxys.append(result)     
    
    return result




##############################################################################
#
# ProxyObject
#
class _ProxyObject(GCommBase.ProxyObject):


  # Server raised an exception...
  # Translate the exception into a local python
  # exception and raise it...
  def _exceptionRaised(self, data):
    if data.isGNUeGenerated():
      name, text = string.split(data.message,': ',1)
      # TODO: _lookup???
      exception = _lookup[name]
      raise exception, text
    else:
      # TODO:
      raise StandardError, data.message


  def _close(self):
    # if it is a dynamic object then close it
    if self._attrPath==[self._attr]:
      self._adapter.runMethod(self._attr+'._close')

