#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001 Free Software Foundation
#
# FILE:
# xmlrpc/ClientAdapter.py
#
# DESCRIPTION:
# Set of classes that implement the XML-RPC client adapter for GNUe Comm.
#
# NOTES:
# Requires py-xmlrpc from  http://py-xmlrpc.sourceforge.net/
#
# Client Parameters:
#
#    url         The complete URI address of the XML-RPC service.
#                e.g, url="https://www.domain.com:9876/service.php"
#    ..or..
#    transport   The service transport (either "http" or "https")
#    host        The hostname or IP address of the service
#    port        The port that the service is located on
#



#
# We provide a client driver...
#
CLIENT = 1      # Client Adapter


from gnue.common.commdrivers import GCommBase
from gnue.common import GComm,GConfig

import string, sys

#
# i18n support
#
import gettext

try:
  gettext.install('gnue',GConfig.getInstalledBase('install_prefix')+'/translations')
except IOError:
  sys.stderr.write('Unable to load locale information falling back to default English language translations.\n')    
  import __builtin__
  __builtin__.__dict__['_'] = lambda string:string

try:

  # xmlrpclib could be in a package, so try both
  try:
    import xmlrpc
  except ImportError:
    from xmlrpc import xmlrpc

except ImportError:
  raise GComm.AdapterInitializationError, \
        _("\nUnable to load xmlrpc.  To use the XML-RPC interface, \n") \
      + _("please install xmlrpc from:\n") \
      + "    http://sourceforge.net/projects/py-xmlrpc " \
      + _("the appropriate DEBIAN package is python-xmlrpc.deb")


##############################################################################
#
# ClientAdapter
#
class ClientAdapter(GCommBase.Client):

  def __init__(self, params):

    self._timeout = 1.0
    
    #
    # check for the correct parameters
    #
    
    try:

      #
      #  test for an url parameter
      # 
      try:
        url = params['url']
        if not len(url):
          raise KeyError

        # split url into server + host + path parts
        import urlparse
        
        transport, netloc, path, params, query, fragment = \
                   urlparse.urlparse(url)
        host, port=split(netloc,':',1)
        
      except KeyError:
        #
        #  transport parameter
        #
        try:
          transport = params['transport']
        except KeyError:
          transport = 'http'

        if (transport!='http'):
           raise GComm.AdapterConfigurationError,  \
                 "py-xmlrpc just support http as a transport. " +\
                 "If you need https please have a look at the " +\
                 "'pw_xmlrpc' adapter."
        #
        #   path parameter
        #
        try:
          path = params['path']
        except KeyError:
          path = '/'

        #
        #   port parameter
        #
        try:
          port = params['port']          
        except KeyError:
          port = 8765
        
          
        #
        #   host parameter
        #
        try:
          host = params['host']
        except KeyError:
          host = 'localhost'

          

      self._url = "%s://%s:%s%s" % \
                  ( transport, params['host'], port,path )
        
      # setLoglevel
      if hasattr(params,'loglevel'):
        xmlrpc.setLogLevel(params['loglevel'])
      else:
        xmlrpc.setLogLevel(0)

      # care for non-integer port settings
      if type(port)!=type(12):
        port=string.atoi(port)

      # add an list to store references to proxys for dyn. objects
      self._objectProxys=[]
      
      # create the real xmlrpc client
      self._client = xmlrpc.client(host,port,path)
      
      # add an error handler ( is disabled because of wrong c implementation)
      # self._client.setOnErr(self.errorHandler)

    except KeyError:
      raise GComm.AdapterConfigurationError,  \
            _("To use XML-RPC, you must either specify a 'url' or\n") +\
            _("a 'host' and 'port' setting.")
    except:
      raise GComm.AdapterInitializationError, \
         _("Unable to initialize the XML-RPC interface at %s") % url


  def request(self, service, params={}):
    self._baseproxy = _ProxyObject(self, service)
    return self._baseproxy


  def errorHandler(self,src, exc):
    # doesn't work correct, so is being disabled
    if exc[1]==(111, 'Connection refused'):
      # cannot raise a new error, so I try to change it
      exc=(GComm.AdapterInitializationError,
           _("Unable to initialize the XML-RPC interface at %s: ")\
           % src.getDesc() +  _("Connection refused"), None )
    print "Connection to ",src.getDesc()," failed."
    print exc
    # sys.exit(0)
    #import traceback
    #sys.stderr.write()
    #traceback.print_exception(exc[0], exc[1], exc[2])
    #return xmlrpc.ONERR_KEEP_WORK
    return xmlrpc.ONERR_KEEP_DEF
  
  def close(self):
    # do a cleanup of the proxy objects and the object stubs
    # on the XMLRPC server side

    # 1. close the proxy objects for services (static, client only)
    self._baseproxy=None    
    
    # 2. close the proxy objects for objects (dynamic, client+server)
    for i in self._objectProxys:
      i._close()      
    self._objectProxys=[] 
    
    # close the client it self
    self._client=None    

  # just in case the user forgot to delete some objects
  def __del__(self):
    if self._client!=None:
      self.close()
    GCommBase.Client.__del__()

  def setTimeout(self,timeout):
    self._timeout=timeout
    
  def setAuthentification(self,name,passw):
    self._auth_name=name
    self._auth_passw=passw


  def runMethod(self, method, *args, **params):
    # print "Calling method: %s with attr %s " % method,string.join(args,",")
    try:
      # send authentication info?
      if hasattr(self,'_auth_name'):
        result = self._client.execute(method, args, self._timeout, \
                                      self._auth_name, self._auth_passw)
      else:
        result = self._client.execute(method, args,self._timeout)
        
    except:      
      if str(sys.exc_type)=="xmlrpc.error":      
        if sys.exc_value==(111, 'Connection refused'):
          msg= _("Unable to connect to XML-RPC server ") + \
               _("at '%s' \n(connection refused)\n") % self._url + \
               _("please check if the server is running")        
          #raise GComm.CommunicationsError, msg
        else:
          msg="Error: %s" % (sys.exc_value,)
          
        print 
        print "********************************************\n"
        print msg,"\n\n"
        print "********************************************"
        sys.exit(0)
        
      elif str(sys.exc_type)=="xmlrpc.fault":
        #
        #  Distinguish between different kinds of distant RPC errors
        #

        # catch unknown service faults created by xmlrpc/ServerAdapter
        if sys.exc_value.faultString[:30]=='xmlrpc.error: unknown command:':
          msg=_("Service/Command %s does not exist on the server")\
               % sys.exc_value.faultString[31:]
          raise GComm.InvalidService,msg,sys.exc_traceback
        
        # catch unknown service faults created by pw_xmlrpc/ServerAdapter  
        elif sys.exc_value.faultString[:26]== \
                 'exceptions.AttributeError:':
          msg=_("%s")\
               % sys.exc_value.faultString[26:]
          raise GComm.InvalidService,msg,sys.exc_traceback

        # catch every other distant error
        else:
          msg =_("Error '%s' occured on server connected"+ \
                 " by XMLRPC") % sys.exc_value.faultString
          raise GComm.DistantError,msg,sys.exc_traceback

      
      else:
        raise GComm.Error, _("Unknown XMLRPC Error: '%s' of type '%s'") % \
              (sys.exc_value, sys.exc_type), \
              sys.exc_traceback
       

      
    # print "the result is",result,"of type",type(result)
    
    # check, if an object handle is send    
    # TODO: make a better check
    if type(result)==type("String") and len(result)==40 :
      
      # create an proxy for this handle      
      result = _ProxyObject(self, "[%s]" % result)

      # store proxy in list
      self._objectProxys.append(result)
    
    return result




##############################################################################
#
# ProxyObject
#
class _ProxyObject(GCommBase.ProxyObject):


  # Server raised an exception...
  # Translate the exception into a local python
  # exception and raise it...
  def _exceptionRaised(self, data):
    if data.isGNUeGenerated():
      name, text = string.split(data.message,': ',1)
      # TODO: _lookup???
      exception = _lookup[name]
      raise exception, text
    else:
      # TODO:
      raise StandardError, data.message

  def _close(self):
    # if it is a dynamic object then close it
    if self._attrPath==[self._attr]:
      self._adapter.runMethod(self._attr+'._close')
