#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001 Free Software Foundation
#
# FILE:
# xmlrpc/ServerAdapter.py
#
# DESCRIPTION:
# Set of classes that implement the XML-RPC server driver for GNUe Comm.
#
# NOTES:
# Requires py-xmlrpc from http://sf.net/projects/python-xmlrpc
# 
#
# Server Parameters:
#
#    port        The port that the service is located on
#



##############################################################################
#
# We provide a server driver...
#
SERVER = 1      # ServerAdapter


#from gnue.common.commdrivers._helpers.AsyncSocketServer import AsyncHTTPServer


from gnue.common import GComm, GDebug
from gnue.common.commdrivers._helpers import ObjectLibrarian
from gnue.common.commdrivers._directory import DirectoryServer
from gnue.common.commdrivers import GCommBase

import string, sys, thread, traceback

try:
  import xmlrpc

except ImportError:
  raise GComm.AdapterInitializationError, \
        _("\nUnable to load xmlrpc.  To use the XML-RPC interface, \n") \
      + _("please install xmlrpc from:\n") \
      + "    http://sourceforge.net/projects/py-xmlrpc" \
      + _("the appropriate DEBIAN package is python-xmlrpc")


# Mapping from GRPC's datatype to XML-RPC datatypes
_datatypeMap = {
  'integer': 'int',
  'string': 'string',
  'boolean': 'boolean',
  'date': 'dateTime.iso8601',
  'number': 'double',
  'base64': 'base64',
  'binary': 'base64'
}


##############################################################################
#
# ServerAdapter
#
class ServerAdapter(DirectoryServer.DirectoryServer):

  def __init__(self, rpcdef, bindings, params):

    ## Check parameters
    try:
      self._port = params['port']
    except KeyError:
      raise GComm.AdapterConfigurationError,\
            _('Required parameter "port" not supplied')

    # Create Subserver
    self.server=XMLRPC_Server()

    
    ## Build up Service Directory
    DirectoryServer.DirectoryServer.__init__(self, rpcdef, \
                                             bindings, params)
    
    if hasattr(params,'loglevel'):
      self._loglevel = params['loglevel']
    else:
      self._loglevel = 1
      
    xmlrpc.setLogLevel(self._loglevel)
    
    self.server.setParent(self)
    self.server.bindAndListen(self._port)
    self.server.setOnErr(self.OnErr)

  def serve(self):
    while 1:
      try:
	self.server.work()
	# you could set a timeout if desired
      except:
        e = sys.exc_info()
        if e[0] in (KeyboardInterrupt, SystemExit):
          raise e[0], e[1], e[2]
        traceback.print_exc()

  def addRpMethod(self,object,parent,binding):    

    DirectoryServer.DirectoryServer.addRpMethod(self,object, \
                                                parent,binding)
    self.server.addMethods({object._path:binding})


  def addRpAttribut(self,object,parent,binding):

    DirectoryServer.DirectoryServer.addRpAttribut(self,object,\
                                                  parent,binding)

    if not object.readonly:
      # Add the set_* directory entry
      self.server.addMethods({'%s.set_%s' % \
                              (parent._path, object.name):\
                              binding})

    if not object.writeonly:
      # Add the get_* directory entry
      self.server.addMethods({'%s.get_%s' % \
                              (parent._path, object.name):\
                              binding})


  def OnErr(server,source):
    print _("Internal server error occured:\n"+
            " server %s \n on source %s.") % (server,source)
    return xmlrpc.ONERR_KEEP_WORK


  #
  # Call the requested method
  #
  def call(self, method, params):
    if self._loglevel>0:
      print _("Dispatching: "), method, params 
    

    ## Check if the Method is part of a service or a pointer to a
    ## single object
    ##
    ## Call to an object:  method="_Management:235423456_.getAttr"
    ##                     params=""
    ## Call to an method: (of a service=one object)
    ##                     method="DonutPlace.Management.Restart"

    if method[0]=='[':
      # call to an object
      # 1. get the object from the objectlibrarian
      # 2. check, if the object is supported by the gfd
      try:
        i=string.index(method,']',1)
        objhandle=method[1:i]
        method=method[i+2:]
      except ValueError:
        raise AttributeError, \
              _("Wrong format of object handler ")+\
              _("in method call %s") % method
      # TODO check in service dir, if obj is supported or not
      o=ObjectLibrarian.retrieveObject(objhandle)
      try:
        server_method=getattr(o,method)
        server_attribute=None
      except AttributeError:
        server_method=None
        try:
          server_attribute=getattr(o,method[4:])
        except AttributeError:
          
          if method!="_close":                      
            msg=_("Internal XMLRPC server error: method %s can be ")% \
                 method +\
                 _("found in the directory (build out of a .grpc file),")+\
                 _(" but the object doesn't contain this method/attribut.")+\
                 _(" Please check you .grpc file for wrong return types.")
            
            raise AttributeError, msg
        
      if method!="_close":
        direntry = self.getMethodDirEntry(o._type+"."+method)
        signature=direntry['signature']
      else:
        signature= ('string',)                

    else:

      # call to a service method or a helper call (get/set) for
      # a service attribut
      try:
        direntry = self.getMethodDirEntry(method)
        server_method = direntry['binding']
        server_attribute = None
        
        # check if it is an real method (binding -> method)
        # or an get/set method for an attribut (binding-> attribut)
        if (type(server_method)!=type(self.call)):
          server_attribute = server_method
          server_method=None
          
        signature=direntry['signature']

        if (server_method==None) and (server_attribute==None):
          raise AttributeError, \
              _("Server XML-RPC method %s  is not ")% method +\
              _("bound to real method") 
      except KeyError:
        raise AttributeError, \
              _("Server does not have XML-RPC ") +\
              _("procedure %s") % method      
    try:
      #
      pass
        # TODO:  Compare submitted attributs with signature
    except KeyError:
      raise AttributeError, \
            _("Server XML-RPC ") +\
            _("procedure %s accepts just %s as attributs") % (method,attr)
    

    # replace object handles in param with the real object
    counter=0
    while counter<len(params):
      p=params[counter]
      if type(p)==type(""):
        if (len(p)==42) and (p[0]=="[") and (p[41]=="]"):
          try:
            p=p[1:41]
            obj=ObjectLibrarian.retrieveObject(p)
            newp=params[0:counter-1]+(obj,)+params[counter+1:]
            params=newp
          except:
            pass
      counter=counter+1;

    # check if it is an real method (binding -> method)
    # or an get/set method for an attribut (binding-> attribut)

    if (server_method!=None):
            
      # call method with params
      result=server_method(*params)

    # check if it's the close object method
    elif (method=="_close"):
      
      result=self.releaseDynamicObject(o)
      
    else:
      
      ## check wether its the set or the get method for the attribut
      mparts=string.splitfields(method,'.')
      mparts.reverse()
      calltype=mparts[0]
      calltype=calltype[:4]
      GDebug.printMesg(7,'method %s has calling type %s' %\
                       (method,calltype))
      if calltype=='set_':
        # setAttribut method
        server_attribute=params[0]
      elif calltype=='get_':
        # getAttribut method
        result=server_attribute
      else:
        raise AttributeError, \
            _("Internal Server XML-RPC error: method type") +\
            _("(get/set attribute) couldn't be detected (method %s)") \
            % method       
    

    # replace real object in param with an object handle
    if type(result)==type(self):  ## both should be instances
       ObjectLibrarian.archiveObject(result)

       # get the type of the result
       rtype=signature[0]
       # delete the surrounding brackets < >
       rtype=rtype[1:len(rtype)-1]
       # store typeinfo in new object
       result._type=rtype

       result=ObjectLibrarian.getObjectReference(result)
       self.registerDynamicObject(result,rtype)

    # check for empty results (not allowed for XMLRPC)
    if result==None:
      GDebug.printMesg(3,'Transform result None into 1')
      result=1

    return result

  def registerDynamicObject(self,objhandle,type):
    #
    #  add new methods to the xmlrpc server
    #
    tl=len(type)
    for i in self.directory.keys():
      if i[0:tl]==type:
        method="["+objhandle+"]"+i[tl:]
        GDebug.printMesg(1,'Method %s registered to py-xmlrpc ' \
                         % method +\
                         ' internal directory.')
        self.server.addMethods({method:None})
     
    # add a method to close this object
    self.server.addMethods({"["+objhandle+"]._close":None})

  def deregisterDynamicObject(self,object):
    #
    #  remove the new methods from the xmlrpc server
    #
    objhandle=ObjectLibrarian.getObjectReference(object)
    type=object._type
    tl=len(type)
    for i in self.directory.keys():
      if i[0:tl]==type:
        method="["+objhandle+"]"+i[tl:]
        GDebug.printMesg(1,'Method %s is deleted from py-xmlrpc ' \
                         % method +\
                         ' internal directory.')
        self.server.removeMethods({method:None})

    # remove the method to close this object
    self.server.removeMethods({"["+objhandle+"]._close":None}) 

  def releaseDynamicObject(self,object):
    # delete bindings for this object in the xmlrpc method directory
    self.deregisterDynamicObject(object)
    # release object for Object Librarian
    ObjectLibrarian.deferenceObject(object)
    

class XMLRPC_Server(xmlrpc.server):
  def setParent(self, parent):
    self._parent=parent

  def dispatch(self, servp, srcp, uri, method, params):
    return self._parent.call(method,params)

  def removeMethods(self, dict):
    d = {}
    for (name, func) in dict.items():
      del self.comtab[name]
      # TODO: remove method from internal comtab too
      # del self._o.comtab[name]
      # or
      # self._o.removeMethods(d)

##############################################################################
#
# Used internally by the XML-RPC server to proxy
# between Object-by-Ref and actual objects.
#
class _ObjectByReferenceHandler:

  def __init__(self, oclass):
    self._class = oclass

  # If this makes absolutely no sense, move on and trust the code :)
  def requestHandler(self, method, *params, **args):

    if len(params):

      # If parameters were passed, the first one is always the handle
      refid = params[0]

      # The object's real method doesn't expect to receive the object handle
      params = params[1:]

    else:

      try:

        # If only named parameters are present, then
        # obj_handle should be the object handle...
        refid = args['obj_handle']

        # The object's real method doesn't expect to receive the object handle
        del args['obj_handle']

      except KeyError:
        raise StandardError, _('Object handle not returned')   # TODO

    try:
      return ObjectLibrarian. \
        retrieveObject(refid).__dict__[method](*params, **args)
    except KeyError:

      # Attempt to use an invalid objecthandle
      raise StandardError, _('Invalid object handle') # TODO

