#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it 
# and/or modify it under the terms of the GNU General Public 
# License as published by the Free Software Foundation; either 
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be 
# useful, but WITHOUT ANY WARRANTY; without even the implied 
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public 
# License along with program; see the file COPYING. If not, 
# write to the Free Software Foundation, Inc., 59 Temple Place 
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000, 2001 Free Software Foundation
#
# FILE:
# appserver/DBdriver.py
#
# DESCRIPTION:
# Implementation of dbdriver for use with GNUe Application Server.
#
# NOTES:
#
# HISTORY:
#

VERSION="0.0.1"

from gnue.common import GDataObjects, GDebug, GConditions, GComm
import string
import types
import md5
import sys

class GEASv2_RecordSet(GDataObjects.RecordSet):
  def setAssoziatedInstance(self,instance):
    self._instance=instance
    
  def _postChanges(self):
    if self._deleteFlag:
      GDebug.printMesg(5, 'AppServer database driver: Instance deleted')
      self._instance.delete()
#      self._instance._close()

    if self._insertFlag:      
      GDebug.printMesg(5, 'AppServer database driver: new Instance '+\
                       'created and inserted')
      self._instance=self._parent._cursor.insertNewInstance()
      
    for field in (self._modifiedFlags.keys()):      
      self._instance.put(field,self._fields[field])
      
    self._updateFlag = 0
    self._insertFlag = 0
    self._deleteFlag = 0
      

  
# GEASv2_ResultSet
#
# Notes:
# In the GEASv2 driver a CURSOR is simply the List handle returned
# via the query interface
#
class GEASv2_ResultSet(GDataObjects.ResultSet): 
  def __init__(self, dataObject, cursor=None, \
        defaultValues={}, masterRecordSet=None): 
    GDataObjects.ResultSet.__init__(
           self,dataObject,cursor,defaultValues,masterRecordSet)
    self._recordSetClass = GEASv2_RecordSet
    
    # Populate field names
    self._fieldNames = self._dataObject.getFields(self._dataObject.table)    

    GDebug.printMesg(5, 'ResultSet created')


  def _loadNextRecord(self): 
    if self._cursor:
   
      # Populate the result set
    
      # load all instances into the cache
      
      Instance=self._cursor.nextInstance()
      dict = {}
        
      # xmlrpc GNURPC driver changes None into 1
      # so we have to check for None or 1
      while (Instance!=None) and (Instance!=1):
        dict = {}
        for fieldName in self._fieldNames:
            dict[fieldName] = Instance.get(fieldName)
            
        record=self._recordSetClass(parent=self,initialData=dict)
        record.setAssoziatedInstance(Instance)
        self._cachedRecords.append (record)

        # increase record count by one
        # (should be replaced by a geasList.getQueryCount() later)
        self._recordCount=self._recordCount+1

        # try to load next instance
        Instance=self._cursor.nextInstance()

      # check if an record has added to the cache
      if dict != {}:
        return 1 

    # if no record returned return a zero
    return 0
    
  # Post changes to the database
  # similar function like in GDataObjects, just with an missing line
  def post(self, foreign_keys={}):
    # post our changes

    recordPosition = 0
    while recordPosition < len(self._cachedRecords):
      record = self._cachedRecords[recordPosition]

      delete = record._emptyFlag or record._deleteFlag
      if not delete:
        # Flip the flag for 'default' values to true so that hidden
        # default fields are included in insert statements
        if record.isPending():
          for field in self._dataObject._defaultValues.keys():
            record._modifiedFlags[field] = 1

        for field in foreign_keys.keys():
          record._fields[field] = foreign_keys[field]
          # Some DBs will throw an exception if you update a Primary Key
          # (even if you are updating to the same value)
          if record._insertFlag:
            record._modifiedFlags[field] = 1

        recordPosition += 1
      else:
        # Adjust the current record if a preceding record
        # or the current record is deleted
        if recordPosition <= self._currentRecord:
          self._currentRecord -= 1
        self._cachedRecords.pop(recordPosition)
        self._recordCount -= 1

      record.post()

    # Move to record 0 if all preceding records were deleted
    # (or set to -1 if all records were deleted)
    if self._currentRecord < 0:
      if len(self._cachedRecords):
        self._currentRecord = 0
      else:
        self._currentRecord = -1


class GEASv2_DataObject(GDataObjects.DataObject):
  # ConditionalName (min args, max args, field creation, bypass function )
  #
  # Commented out conditionals are a bad thing
  conditionElements = {
       'add':             (2, 999, None,                   'None'      ),
       'sub':             (2, 999, None,                   'None'      ),
       'mul':             (2, 999, None,                   'None'      ),
       'div':             (2, 999, None,                   'None'      ),
       'and':             (1, 999, None,                   'None'      ),
       'or':              (2, 999, None,                   'None'      ),
       'not':             (1,   1, None,                   'None'      ),
       'negate':          (1,   1, None,                   'None'      ),
       'eq':              (2,   2, '(%s == %s)',                None     ),
       'ne':              (2,   2, '(%s != %s)',             None     ),
       'gt':              (2,   2, '(%s > %s)',              None     ),
       'ge':              (2,   2, '(%s >= %s)',                None     ),
       'lt':              (2,   2, '(%s < %s)',                 None     ),
       'le':              (2,   2, '(%s <= %s)',                None     ),
       'like':            (2,   2, '(%s LIKE %s)',               None     ),
       'notlike':         (2,   2, '%s NOT LIKE %s',            None     ),
       'between':         (3,   3, '%s BETWEEN %s AND %s',      None     )
       }

  def __init__(self):
    GDataObjects.DataObject.__init__(self)

    GDebug.printMesg (1,"AppServer database driver backend initializing")
    self._resultSetClass = GEASv2_ResultSet
    self._DatabaseError = GComm.Error

  def connect(self, connectData={}): 
    GDebug.printMesg(1,"AppServer database driver connecting...")
    # Extract the username and password...
    user = connectData['_username']
    passwd = connectData['_password']

    params = { 'host': connectData['host'],
               'port': connectData['port'],
               'transport': connectData['transport']}
    
    self._server = GComm.attach( connectData['rpctype'],params)
    
    GDebug.printMesg(3,"Setup an link to the session manager")
    sessionManager = self._server.request("SessionManager")
    
    GDebug.printMesg(3,"Get the status of the session manager")
    GDebug.printMesg(3,"Status: "+sessionManager.Status())
    
    GDebug.printMesg(3,"Creating session object ...")  
    self._session=sessionManager.getNewSession()

    try:
      GDebug.printMesg(1,"Logging into appserver as user '%s'" % (user))
      if (user == 'guest') or (user == ''):

        # TODO: (jcater) I'm not so sure about guest logins. This
        # TODO: (jcater) should be discussed. At a minimum, I think
        # TODO: (jcater) we need gnue.conf flags:
        # TODO: (jcater)     allowGuestLogin = [0|1]
        # TODO: (jcater)     guestLogin = guest
        # TODO: (jcater)     allowBlankLogin = [0|1]
        if not self._session.login(user,passwd):
          # self._session.close()
          raise GDataObjects.ConnectionError, "Error loging into appserver"
      else:

        # TODO: (jcater) The GEAS docs say this is not supported and
        # TODO: (jcater) CORBA-based security should instead be used.
        # TODO: (jcater) As I don't know jack about CORBA security,
        # TODO: (jcater) this will have to do for now (even though,
        # TODO: (jcater) IIRC, is does nothing to validate the user/pass
        # TODO: (jcater) supplied to it.)
        if not self._session.login(user,passwd):
          # self._session.close()
          raise GDataObjects.ConnectionError, "Error loging into appserver"

#      self.triggerExtensions = TriggerExtensions(self._dataConnection)
      self.triggerExtensions = TriggerExtensions(self._session)

    except:
      raise GDataObjects.ConnectionError, "Error loging into appserver"
  
 

    #
    # Enter transactional mode...
    #   TODO: (jcater) As transaction support does not yet exist in GEAS, 
    #   TODO: (jcater) I have no idea if this works, if it will change, 
    #   TODO: (jcater) or, well, anything for that matter. Think of it  
    #   TODO: (jcater) as a placeholder.
#    self._transaction = self._dataConnection.new()


  # We only need the basics -- username and password -- to log in
  def getLoginFields(self): 
    return [['_username', 'User Name',0],['_password', 'Password',1]]

  def _buildQuery(self, conditions={}):
    # Standardize incomming conditions as a GConditions structre
    if type(conditions) == types.DictType:
        cond = GConditions.buildConditionFromDict(conditions)
    else:
        cond = conditions

    # Construct query object
    GDebug.printMesg(7,'Implicit Fields: %s' % self._fieldReferences)
    query = []
    
    # Add conditionals
    query = GConditions.buildPrefixFromTree(cond._children[0] )
    
    GDebug.printMesg(7,'Full query in prefix notation: %s' % query)
    
    return query

                      
  def _createEmptyResultSet(self, readOnly=0, masterRecordSet=None):
    return self.createResultSet(readOnly=readOnly,\
                                conditions=GConditions.GCimpossible,\
                                masterRecordSet=masterRecordSet)

  def _createResultSet(self, conditions={}, readOnly=0, masterRecordSet=None):
    try:
        self._list = self._session.createList (self.table)
    except:
        raise GDataObjects.ConnectionError, "Error during creation of object list"
  
    GDebug.printMesg (5,"Setting up list object ...");

    # the fieldnames seem to be stored in _fieldReferences
    fieldlist=self._fieldReferences.keys()
    self._list.setPrefetch (fieldlist);
    
    GDebug.printMesg (5,"Using "+str(fieldlist)+" as fieldlist");

    if hasattr(self,"order_by"):
      GDebug.printMesg (5,"Sorting list by fields '"+self.order_by+"'");
      sort=string.splitfields(self.order_by,",")
      self._list.setSort (sort)

    try:
      if conditions:
        GDebug.printMesg (5,"Setting Conditions ...");
        self._list.setConditions(self._buildQuery(conditions))      
    except self._DatabaseError, err: 
      raise GDataObjects.ConnectionError, err
  

    GDebug.printMesg (5,"Populating list ...");
    self._list.populate ()
    
    rs = self._resultSetClass(self, cursor=self._list, masterRecordSet=None)
    
    if readOnly: 
      rs._readonly = readOnly
    return rs

  def commit(self): 
    GDebug.printMesg (5,"AppServer database driver: commit()")
 
    try:
        # at the moment there is only commit on session layer
        self._session.commit()
    except self._DatabaseError, value:
      raise GDataObjects.ConnectionError, value

  def rollback(self): 
    GDebug.printMesg (5,"AppServer database driver: rollback()")

    try:
      # at the moment there is only rollback on session layer
      self._session.rollback()      
    except: 
      raise GDataObjects.ConnectionError, "Error during Rollback"	

### The folowing methods are not working and just copies from the old geas
### driver

### TODO add introspection support to geas

  def getSchemaTypes(self):
      return [('object','Object',1)]

  # Return a list of Schema objects
  def getSchemaList(self, type=None):
      includeObjects = (type in ('object','sources', None))
      
      list = []
      for classname in self._dataConnection.classes:
        list.append(self.getSchemaByName(classname))
      return list

  def getSchemaByName(self, name, type=None): 
      classdef = self._dataConnection.getFullClassDefinition(str(name))
      schema = GDataObjects.Schema(attrs={'id':string.lower(classdef.name),
                                          'name':classdef.name,
                                          'type':'object'},
                                   getChildSchema=self.__getFieldSchema)
      return schema
  
  # Get fields/methods of a GEAS object
  def __getChildSchema(self, parent):
      list = []
      c = con.getFullClassDefinition( classnames[idx-1] )
      
  def getFields(self, name):
      # no introspection support for now
      return self._fieldReferences.keys()

supportedDataObjects = { 
  'object': GEASv2_DataObject
}


class TriggerExtensions:

    def __init__(self, connection):
        pass
