#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000, 2001 Free Software Foundation
#
# FILE:
# GFClient.py
#
# DESCRIPTION:
# Client startup file that parses args, builds form from file
# and passes control to a UI.
#
# NOTES:
#  Once all the events are moved back in here
#    debug
#    make the next/prec methods more generic in the GFForm
#    change self._form in here to a list

import os, copy
import os.path
import sys
import urllib

from gnue.common.GClientApp import *
from gnue.forms.GFInstance import *
from gnue.forms.GFForm import *
from gnue.forms.GFParser import loadForm
from gnue.forms import GFKeyMapper
from gnue.common import GDebug
from gnue.common import GConfig, GDataObjects, GConnections, dyn_import
from gnue.forms import VERSION
from gnue.forms.GFConfig import ConfigOptions

class GFClient(GClientApp):
  #
  # GClientApp() overrides
  #
  VERSION = VERSION
  COMMAND = "gnue-forms"
  NAME = "GNUe Forms"
  USAGE = GClientApp.USAGE + " file"
  COMMAND_OPTIONS = [
      [ 'user_interface', 'u', 'user_interface', 1, 'wx', 'type',
          _('The currently supported values for <type> are ') \
          +'wx(GUI), nttsi(curses)' ],
      [ 'splash_screen','s','no_splash', 0, None, None,
          _('Disables the splash screen')]
      ]
  SUMMARY = \
     _("GNUe Forms is the primary user interface to the GNU Enterprise system.")

  #
  # Initialize the class
  #
  def __init__(self, connections=None):
    GClientApp.__init__(self, connections,'forms',ConfigOptions)

    self._formInstances = {}
    self._lastSerialNumber = 0

    self.ui_type = self.OPTIONS['user_interface']

    if gConfig('disableSplash') == '1': 
      self.disableSplash = 1
    else:
      self.disableSplash = self.OPTIONS['splash_screen']

  #
  # Run the client application
  #
  # GClientApp().run() override
  #
  def run(self):

    # assign form file from 1st free argument
    try:
      formfile = self.ARGUMENTS[0]
    except:
      try:
        # If no form specified, then see if this is a symlinked form definition
        if string.split(string.lower(os.path.basename(sys.argv[0])),'.')[0] != 'gnue-forms':
          formfile = os.path.join (
              GConfig.getInstalledBase('forms_appbase','common_appbase','install_prefix'),
              gConfig('FormDir'),
              os.path.basename(sys.argv[0])+".gfd" )
        else:
          raise IndexError
      except IndexError:
        self.handleStartupError (_("No Forms Definition File Specified."))


    #
    # Get the user supplied parameters
    #
    paramList = self.ARGUMENTS[1:]

    userParameters = {}

    for param in paramList:
      psplit = string.split(param,'=',1)
      if len(psplit) == 1:
        self.handleStartupError ( \
            'Parameter "%s" specified, but no value supplied.' % psplit[0] )
      userParameters[string.lower(psplit[0])] = psplit[1]

      GDebug.printMesg(2,'Param "%s"="%s" ' % \
           (string.lower(psplit[0]), psplit[1]))


    self.loadui ()

    self.loadKeyMappings()
    self.runForm(formfile, self.disableSplash, parameters=userParameters)


  def loadui (self):
    # Initialize user interface (defaults to gui)
    #  If the DISPLAY var isn't set then switch to text
    uiloaded = 0
    while not uiloaded:
      try:
        self._ui = dyn_import("gnue.forms.uidrivers.%s" % (self.ui_type))
        uiloaded = 1
      except ImportError, err:
        GDebug.printMesg(1,_("Unable to load UI driver")+" %s" % self.ui_type)
        GDebug.printMesg(1,err)
        if self.ui_type != 'nstti':
          self.ui_type = 'nstti'
        else:
	  self.handleStartupError(_("Unable to load any valid UI drivers.  Aborting."))
          sys.exit()
  

  def loadKeyMappings(self):
    options = gConfigDict()
    mapping = {}

    for key in options.keys():
      if string.lower(key)[:4] == 'key_':
        mapping[key[4:]] = options[key]

    GFKeyMapper.KeyMapper.loadUserKeyMap(mapping)


  def runFormFromTrigger(self, formFile, parameters={}):
    self.runForm(formFile, disableSplash=1, parameters=parameters)


  def runForm(self, formFile, disableSplash=0, parameters={}):
    """
      test
    """
    try:
      #
      # Create the instance
      #
      instance = GFInstance(self, self.getNextSerialNumber(),
          connections=self.connections, ui=self._ui, disableSplash=disableSplash,
          parameters=parameters)
      self._formInstances[instance.getSerialNumber()] = instance

      #
      # Assign the proper login handler based upon the user interface choice
      #
      self.getConnectionManager().setLoginHandler(self._ui.UILoginHandler())

      #
      # Build the form tree
      #
      drive = os.path.splitdrive(formFile)
      if len(drive[0]):
        fileHandle = open(formFile,'r')
      else:
        fileHandle = urllib.urlopen(formFile)

      form = loadForm (fileHandle, instance, 1, parameters)
      fileHandle.close()

      #
      # Start the instance
      #
      instance.setForm(form)
      instance.activate()

    except GConnections.Error, mesg:
      self.handleStartupError(mesg)

    except IOError, mesg:
      self.handleStartupError(_("Unable to open file\n\n     %s")%mesg)

    except GDataObjects.ConnectError, mesg:
      self.handleStartupError(\
         _("Unable to login to datasource.\n\n       %s") %mesg)

    except GDataObjects.ConnectionError, mesg:
      self.handleStartupError(\
         _("Error while communicating with datasource.\n\n       %s") %mesg)

    except GDataObjects.Error, mesg:
      self.handleStartupError(mesg)


  def getNextSerialNumber (self):
     self._lastSerialNumber = self._lastSerialNumber + 1
     return self._lastSerialNumber

if __name__ == '__main__':
  GFClient().run()


