from __future__ import nested_scopes

#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GFForm.py
#
# DESCRIPTION:
# Class that contains the internal python object representation of a GNUE Form
# built from GFObjects
#
# NOTES:
#


import sys
import string
import traceback

from gnue.common import GDebug, openResource, GConnections
from GFObjects import *
from GFEvent import *
from gnue.common.GDataObjects import ConnectionError as DBError
from gnue.common.GRootObj import GRootObj
#from GFTriggerError import *
from gnue.common.GTrigger import TriggerError
from GFLibrary import *
import GFLibrary
import GFParser
import GFTrigger

from gnue.common import GDebug

# Defines which objects are "Tab Stops"
TabStops = ('GFEntry','GFButton')

class GFMainMenuBar(GFObj, GFEventAware):
  def __init__(self, parent=None, window=None):
    GFObj.__init__(self, parent)
    self._type = "GFMainMenuBar"
    self._currentObject = [self]

class GFMainToolBar(GFObj, GFEventAware):
  def __init__(self, parent=None, window=None):
    GFObj.__init__(self, parent)
    self._type = "GFMainToolBar"
    self._currentObject = [self]


class GFForm(GFObj, GRootObj, GFEventAware):
  def __init__(self, parent=None, app=None):
    GRootObj.__init__(self)
    GFObj.__init__(self, parent)
    self._type = "GFForm"
    self._currentObject = [self]

    self._databaseDictionary = {}
    self._datasourceDictionary = {}

    # Dictionary to locate named triggers
    self._triggerDictionary = {}

    # Insert/Overwrite mode
    self._insertMode = 1

    # Focus information
    self._currentPage = None
    self._currentBlock = None
    self._currentEntry = None

    self._blockList = []
    self._pageList = []

    self._app = app

    self._triggerns={}

    # This will be set by GFInstance
    self._parameters = {}

    # The "None" init gives datasources time to setup master/detail
    self._inits = [self.primaryInit, None, self.secondaryInit]

    #
    # New trigger support
    #
    self._triggerGlobal = 1
    self._triggerFunctions = {'setFocus':{'function':self.triggerSetFocus,
                                          'global': 1,
                                          },
                              'setStatusText':{'function':self.triggerSetStatusText,
                                          'global': 1,
                                          },
                              'genericBox':{'function':self.triggerGenericBox},
                              'callTrigger':{'function':self.triggerCallTrigger, 'global': 1},
                              'exit':{'function':self.exit},
                              'commit':{'function':self.commit},
                              'postChanges':{'function':self.postChanges},
                              'reset':{'function':self.reset},
                              'clear':{'function':self.clear},
                              'commitConnection':{'function':self.commitConnection},
                              'rollbackConnection':{'function':self.rollbackConnection},
                              'setParameter':{'function':self.setParameter},
                              'getParameter':{'function':self.getParameter,
                                          'global': 1,
                                          },
                              }



  def objectIsFocusable (self, object):
    """
      true if object is focusable (duh!)
    """
    return object._type=="GFButton" or (object._type=="GFEntry" and not object.style=="label" and not object.hidden and not object.style=='textlookup')

  #
  # Convert some deprecated options to new style
  #
  def _buildObject(self):
    i = 0
    while i < len(self._children):
      if isinstance(self._children[i], GFOptions):
        j = 0
        while j < len(self._children[i]._children):
          if self._children[i]._children[j].name in \
             ('title','width','height'):
            self.__dict__[self._children[i]._children[j].name] = \
                 self._children[i]._children[j].value
            self._children[i]._children.pop(j)
          else:
            j = j + 1
        if not len (self._children[i]._children):
          self._children.pop(i)
        else:
          i = i + 1
      else:
        i = i + 1

    return GFObj._buildObject(self)

  #
  # primaryInit
  #
  # Called during phaseInit startup by GParser.
  #
  def primaryInit(self, params={}):

    # Initialize our events system
    GFEventAware.__init__(self, self._app.eventController)


    # TODO: Look at moving this into GParser
    # Import all needed objects
    self.walk(self.initImportedObjects)

    # analyze the tree
    self.walk(self.analyzeTree)

    self.initTriggerSystem()
    self._triggerns.update(self._triggerNamespaceTree._globalNamespace)
    self._triggerns['runForm'] = self._app.manager.runFormFromTrigger
    self._parameters= params

  def secondaryInit(self, params={}):
    # create the first records
    for key in self._datasourceDictionary.keys():
      if not self._datasourceDictionary[key].hasMaster():
        self._datasourceDictionary[key].createEmptyResultSet()
    self.processTrigger('On-Startup')

  #
  # begin routines made for walking
  #
  # routines that are used by the walk() method
  # suggesting a better way probably exists :)
  #

  #
  # analyzeTree
  #
  # Scans the tree form tree and performs the following
  #
  # builds the various entity lists
  # it also sets the initial focus of the form
  # constructs the trigger namespace
  #
  def analyzeTree(self,object):
    if object._type == 'GFBlock':
      self._blockList.append(object)
      if self._currentBlock == None:
        self._currentBlock = object

    elif object._type == 'GFPage':
      self._pageList.append(object)
      if self._currentPage == None:
        self._currentPage = object

    elif not self._currentEntry and self.objectIsFocusable (object):
      self._currentEntry = object
      GDebug.printMesg (1, "set curreent entry to %s" % object.name)

  def getParameter(self, parameter):
    GDebug.printMesg (1, "Looking for parameter %s" % parameter)
    param = string.lower(parameter)
    try:
      return self._parameters[param]
    except KeyError:
      rv = None
      for child in self._children:
        if isinstance(child, GFParameter) and child._name == param:
          rv = child.default
          self._parameters[param] = rv
          break
    return rv

  def setParameter(self, parameter, value):
    GDebug.printMesg (1, "Looking for parameter %s" % parameter)
    param = string.lower(parameter)
    self._parameters[param] = value

  #
  # Find focus
  #
  # Find the next focusable item given
  # an item of type Page, Block, or Entry
  #
  def findFocus(self, object):
    entry = None

    if isinstance(object, GFObj):
      if self.objectIsFocusable (object):
        return object
      else:
        for child in object._children:
          entry = self.findFocus(child)
          if entry:
            break

    return entry

  #
  # findAndChangeFocus
  #
  # Runs findFocus and then runs
  # changeFocus with that value.
  #
  def findAndChangeFocus(self, object):
    entry = self.findFocus(object)
    if entry:
      self.changeFocus(entry)
    else:
      # print 'nothing'
      pass


  #
  # changeFocus
  #
  # changes to the requested entry object requested by
  # an event source
  #
  def changeFocus(self, widget, fireFocusTriggers=1):

    GDebug.printMesg (1, 'changing focus')
    # for i in dir (widget):
      # attr= getattr (widget, i)
      # if not callable (attr):
        # GDebug.printMesg (1, "widget[%s]= %s" % (i, attr))

    # mdione: como te borraria...
    if not hasattr(widget,'_block') or widget._block==None:
      return None

    message = None

    try:
      if self._currentEntry:
        event = GFEvent('endEDITMODE', None)
        self._app.dispatchEvent(event)
        if event.__error__:
          return 1

      # fieldChange = widget != self._currentEntry
      fieldChange = 1
      blockChange = widget._block != self._currentBlock
      pageChange = widget._block._page != self._currentPage

      if fireFocusTriggers:
        try:
          if fieldChange:
            self._currentEntry.processTrigger('Pre-FocusOut')
          if blockChange:
            self._currentBlock.processTrigger('Pre-FocusOut')
          if pageChange:
            self._currentPage.processTrigger('Pre-FocusOut')
        except AttributeError:
          pass

      oldEntry = self._currentEntry
      oldBlock = self._currentBlock
      oldPage = self._currentPage
      
      self._currentEntry = widget
      self._currentBlock = self._currentEntry._block
      self._currentPage = self._currentBlock._page

      if pageChange:
        self.dispatchEvent(GFEvent('gotoPAGE',self._currentPage));

      if blockChange:
        self.refreshDisplay(self._currentBlock)

      self.dispatchEvent(GFEvent('updateENTRY', oldEntry))
      self.dispatchEvent(GFEvent('updateENTRY', self._currentEntry))

      if fireFocusTriggers:
        try:
          if pageChange:
            oldPage.processTrigger('Post-FocusOut')
          if blockChange:
            oldBlock.processTrigger('Post-FocusOut')
          if fieldChange:
            oldEntry.processTrigger('Post-FocusOut')
        except AttributeError:
          pass

      if fireFocusTriggers:
        if pageChange:
          self._currentPage.processTrigger('Pre-FocusIn')
        if blockChange:
          self._currentBlock.processTrigger('Pre-FocusIn')
        if fieldChange:
          self._currentEntry.processTrigger('Pre-FocusIn')
          self._currentEntry.processTrigger('Post-FocusIn')
        if blockChange:
          self._currentBlock.processTrigger('Post-FocusIn')
        if pageChange:
          self._currentPage.processTrigger('Post-FocusIn')

      event = GFEvent('beginEDITMODE', None)
      self.dispatchEvent(event)
      if event.__error__:
        return 1
    except TriggerError, t:
      GDebug.printMesg(1, "Trigger Error!")
      message = _("Trigger Exception :\n") + t.msg
    return message


  def initImportedObjects(self, object):
    if isinstance(object, GFLibrary.GFImport):
      importAllDataSources = object.datasources == "*"
      importAllPages = object.pages == "*"
      importAllTriggers = object.triggers == "*"
      importDataSources = string.split(string.replace(object.datasources,' ',''),',')
      importPages = string.split(string.replace(object.pages,' ',''),',')
      importTriggers = string.split(string.replace(object.triggers,' ',''),',')

      handle = openResource(object.library)
      form = GFParser.loadForm(handle, self._app, initialize=0)
      handle.close()

      if importAllDataSources or len(importDataSources):
        for child in form._children:
          if isinstance(child, GFDataSource) and \
             (importAllDataSources or child.name in importDataSources):
            child._parent = object
            child._IMPORTED = 1
            object._children.append(child)
            child._buildObject()

      if importAllPages or len(importPages):
        for child in form._children:
          if isinstance(child, GFPage) and \
             (importAllPages or child.name in importPages):
            child._parent = object
            child._IMPORTED = 1
            object._children.append(child)
            child._buildObject()

      if importAllTriggers or len(importTriggers):
        for child in form._children:
          if isinstance(child, GFTrigger.GFTrigger) and \
             (importAllTriggers or child.name in importTriggers):
            child._parent = object
            child._IMPORTED = 1
            object._children.append(child)
            child._buildObject()


    elif isinstance(object, GFLibrary.GFImportItem):

      handle = openResource(object.library)
      form = GFParser.loadForm(handle, self._app, initialize=0)
      handle.close()
      id = 'id'
      if hasattr(object,'name'):
        id = 'name'

      rv = self.__findImportItem(object, form, id)
      if rv != None:
        rv._parent = object
        rv._IMPORTED = 1
        object._children.append(rv)
        for key in object._loadedxmlattrs.keys():
          if key[0] != '_':
            rv.__dict__[key] = object._loadedxmlattrs[key]
            GDebug.printMesg (5, ">>> Moving %s" % key)
        rv._buildObject()

  def __findImportItem(self, find, object, id):
    if isinstance(object, find._importclass) and \
       hasattr(object, id) and \
       object.__dict__[id] == find.__dict__[id]:
      return object
    elif hasattr(object,'_children'):
      rv = None
      for child in object._children:
        rv = self.__findImportItem(find, child, id)
        if rv:
          break
      return rv
    else:
      return None

  #
  # end of routines made for walking
  #


  def updateUIEntry(self,entry):
    self._app.dispatchEvent(GFEvent('updateENTRY',entry));

  #
  # Incomming Event handlers
  #

  def newRecord(self):
    if not self.readonly:
      self._currentBlock.newRecord()

  def deleteRecord(self):
    if not self.readonly:
      self._currentBlock.deleteRecord()

  def isSaved(self):
    saved = 1
    for block in self._blockList:
      if not block.isSaved():
        saved = 0
        break
    return saved

  def prepQuery(self):
    message = None
    if self._currentBlock.mode != 'query':
      for block in self._blockList:
        if not block.isSaved():
          message = _('Data not saved. Save changes or clear the form to proceed.')
          return message

    for block in self._blockList:
      block.processRollback()
      for entry in block._entryList:
        if hasattr(entry,'queryDefault'):
          GDebug.printMesg(1, "%s will be set to %s" % (entry.name, entry.queryDefault))
          entry.setValue(entry.queryDefault)
    self._currentBlock.initQuery()
    return message

  def executeQuery(self):
    self._app.dispatchEvent(GFEvent('beginWAIT',None));

    message = None
    try:
      self._currentBlock.processQuery()

    except DBError:
      self.rollback(1)
      message = _("Database query error:\n%s\n%s ") % (sys.exc_info()[0], sys.exc_info()[1])

    self._app.dispatchEvent(GFEvent('endWAIT',None));

  def commitConnection(self):
    self._currentBlock._dataSourceLink._dataObject.commit()

  def rollbackConnection(self):
    self._currentBlock._dataSourceLink._dataObject.rollback()

  def commit(self, onlyThisBlock=None):
    self._app.dispatchEvent(GFEvent('beginWAIT',None));

    message = None

    try:
      if not self.readonly:
        # Form level pre-commit
        try:
          self.processTrigger('Pre-Commit')
        except TriggerError:
          GDebug.printMesg(1, "Trigger form Pre-Commit threw a TriggerError!")
          self._app.dispatchEvent(GFEvent('endWAIT',None));
          return _("Form trigger returned error")

        blockList = []
        if onlyThisBlock:
          blockList.append(onlyThisBlock)
        else:
          blockList = self._blockList

        for block in blockList:
          GDebug.printMesg(1, "Saving %s"%block.name)
          try:

            # This gets lost in the Pre-Commit code
            block._precommitRecord = block._currentRecord

            # TODO: Need to check with jamest..
            # Now, block-level pre-commits get called for
            # EACH record via another hook in GFDataSource
            ##block.processTrigger('Pre-Commit')

            block.processCommit()
          except TriggerError:
            GDebug.printMesg(1, "Trigger block Pre-Commit threw a TriggerError!")
            self._app.dispatchEvent(GFEvent('endWAIT',None));
            return _("Block trigger returned error")
          block.processTrigger('Post-Commit')

        for block in blockList:
            block.switchRecord(0)

      else:
        message = _('Form is readonly')
        self._app.dispatchEvent(GFEvent('endWAIT',None));
        return message

      self.processTrigger('Post-Commit')

    except DBError:
      self.rollback(1)
      message = _("Database commit error:\n%s\n%s ") % (sys.exc_info()[0], sys.exc_info()[1])
    except:
      print _("\n\nGFForm: Unexpected Exception:")
      print '-'*60
      traceback.print_exc(file=sys.stdout)
      print '-'*60

    self._app.dispatchEvent(GFEvent('endWAIT',None));

    return message

  def postChanges(self, onlyThisBlock=None):
    self._app.dispatchEvent(GFEvent('beginWAIT',None));

    message = None

    try:
      if not self.readonly:
        # Form level pre-commit
        try:
          self.processTrigger('Pre-Commit')
        except TriggerError:
          GDebug.printMesg(1, "Trigger form Pre-Commit threw a TriggerError!")
          self._app.dispatchEvent(GFEvent('endWAIT',None));
          return _("Form trigger returned error")

        blockList = []
        if onlyThisBlock:
          blockList.append(onlyThisBlock)
        else:
          blockList = self._blockList

        for block in blockList:
          GDebug.printMesg(1, "Saving %s"%block.name)
          try:

            # This gets lost in the Pre-Commit code
            block._precommitRecord = block._currentRecord

            # TODO: Need to check with jamest..
            # Now, block-level pre-commits get called for
            # EACH record via another hook in GFDataSource
            ##block.processTrigger('Pre-Commit')

            block.processPostChanges()
            self._currentBlock = block
          except TriggerError:
            GDebug.printMesg(1, "Trigger block Pre-Commit threw a TriggerError!")
            self._app.dispatchEvent(GFEvent('endWAIT',None));
            return _("Block trigger returned error")
          block.processTrigger('Post-Commit')

        for block in blockList:
            block.switchRecord(0)

      else:
        message = _('Form is readonly')
        self._app.dispatchEvent(GFEvent('endWAIT',None));
        return message

      self.processTrigger('Post-Commit')

    except DBError, err:
      #self.rollback(1)
      message = _("Database commit error:\n%s\n%s ") % (sys.exc_info()[0], sys.exc_info()[1])
      #print message
      self._app.dispatchEvent(GFEvent('endWAIT',None));
      raise DBError(str(err))
    except:
      print _("\n\nGFForm: Unexpected Exception:")
      print '-'*60
      traceback.print_exc(file=sys.stdout)
      print '-'*60

    self._app.dispatchEvent(GFEvent('endWAIT',None));

    return message

  def executeAbout(self,event):
     version = VERSION
     name = self.getOption('title') or "Unknown"
     formversion = self.getOption('version') or "Unknown"
     author = self.getOption('author') or "Unknown"
     description = self.getOption('description') or "Unknown"

     self.dispatchEvent(GFEvent('uiABOUT',[version,name,formversion,author,description]))

  def reset(self):
    self.rollback(1)

  def clear(self):
    self.rollback()

  def rollback(self, recover=0):
    #commited = 0
    self.endEditing()
    for block in self._blockList:
      block.processRollback(recover)
      #if not commited \
      #   and str(block._dataSourceLink._dataObject).startswith("<gnue.common.dbdrivers.psycopg.DBdriver.PG_DataObject_Objec"):
      #  block._dataSourceLink._dataObject.rollback()
      #  commited = 1
    self.refreshDisplay(self)
    self._currentBlock.jumpRecord(self._currentBlock._currentRecord)


  #
  # fireTrigger
  #
  def fireTrigger(self, triggerName):
    self._triggerDictionary[triggerName](self)

#
# Updated methods for the new gfclient code
#

  # ugh!
  def gimmeNames (self, x):
    try:
      r= x.name
    except AttributeError:
      try:
        r= x.label
      except AttributeError:
        try:
          r= x._value
        except AttributeError:
          r= x
    return r

  def sortByYAndThenByX (self, a, b):
    # 2= x, 1= y
    if a.yabs<b.yabs:
      return -1
    elif a.yabs==b.yabs:
      if a.xabs<b.xabs:
        return -1
      elif a.xabs==b.xabs:
        return 0
      elif a.xabs>b.xabs:
        return 1
    elif a.yabs>b.yabs:
      return 1

  def absolutize (self, entry, xoff, yoff):
    try:
      entry.xabs= entry.x+xoff
      entry.yabs= entry.y+yoff
    except:
      entry.xabs= xoff
      entry.yabs= yoff
    return entry

  def _getFocusableEntries (self, children, xoff, yoff):
    entryNo= 0
    children= map (lambda x: self.absolutize (x, xoff, yoff), children)
    result= children
    for entry in children:
      # shouldn't this be 'in containers'?
      if entry._type in [ 'GFBox', 'GFBlock' ]:
        next= self._getFocusableEntries (entry._children, entry.xabs, entry.yabs)
        # replace the box with its children recursively
        result= result[:entryNo]+next+result[entryNo+1:]
        entryNo= entryNo+len (next)
      entryNo= entryNo+1

    # GDebug.printMesg (1, "_gFE: result: %s" % map (self.gimmeNames, result))
    result= filter (self.objectIsFocusable, result)
    result.sort (self.sortByYAndThenByX)
#    print map (
#      lambda x:
#        "%s, %d:%d, %d:%d" % (x.name, x.xabs, x.x, x.yabs, x.y)
#      , result
#    )
    return result

  #
  # nextEntry
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry object
  #
  def nextEntry(self):
    firstEntry= self._currentEntry
    nextEntry = None
    keepNext = 0

    entryList= self._getFocusableEntries (self._currentPage._children, 0, 0)
    # print entryList
    # print TabStops
    for object in entryList:
      # Put the first field as the next to rollover
      if self.objectIsFocusable (object):
        if nextEntry == None:
          nextEntry = object
          firstEntry = object

        if object == self._currentEntry:
          keepNext = 1
          continue

        if keepNext == 1:
          nextEntry = object
          break

    self.changeFocus(nextEntry)


  #
  # previousEntry
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry object
  #
  def previousEntry(self):
    # find last entry widget
    nextEntry = None
    lastEntry = self._currentEntry

    entryList= self._getFocusableEntries (self._currentPage._children, 0, 0)
    for object in entryList:
      if self.objectIsFocusable (object):
        nextEntry = object
        lastEntry = object

    keepNext = 0
    for object in entryList:
      # Put the first field as the next to rollover
      if self.objectIsFocusable (object):
        if object == self._currentEntry:
          break

        nextEntry = object

    if self._currentBlock.transparentBlock and nextEntry == lastEntry: # Jump to the prev block if block is tagged as transparent
      prevBlock = self.findPreviousBlock()
      # Move to the new last record of the new current block
      for object in prevBlock._children:
        if self.objectIsFocusable (object):
          nextEntry = object

    self.findAndChangeFocus(nextEntry)


  #
  # refreshDisplay
  #
  # Refreshes all the UI fields to match
  # the virtual form's fields
  #
  def refreshDisplay(self, block):
    block.walk(self.__refreshDisplay)

  def __refreshDisplay(self, object):
    if object._type == 'GFEntry':
      self.dispatchEvent(GFEvent('updateENTRY',object))



  #
  # nextBlock
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry block
  #
  def nextBlock(self):

    try:
      nextBlock = self._blockList[self._blockList.index(self._currentBlock)+1]
    except IndexError:
      nextBlock = self._blockList[0]

    self.findAndChangeFocus(nextBlock)


  #
  # findPreviousBlock
  #
  # Convenience routine to find the previous block
  #
  def findPreviousBlock(self):
    try:
      return self._blockList[self._blockList.index(self._currentBlock)-1]
    except IndexError:
      return self._blockList[-1]


  #
  # previousBlock
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry block
  #
  def previousBlock(self):
    self.findAndChangeFocus(self.findPreviousBlock())


  #
  # endEditing
  #
  # Signal to the current Entry to stop editing
  # mode and save it's value to the virtual form
  #
  def endEditing(self):
    event = GFEvent('endEDITMODE',None)
    self.dispatchEvent(event)
    return not event.__error__

  def beginEditing(self):
    event = GFEvent('beginEDITMODE',None)
    self.dispatchEvent(event)
    return not event.__error__


  def prevRecord(self,block=None):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot go to previous record: in query mode!')
      return
    if block:
      block.prevRecord()
    else:
      self._currentBlock.prevRecord()

  def firstRecord(self):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot go to first record: in query mode!')
      return
    self._currentBlock.firstRecord()

  def lastRecord(self):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot go to last record: in query mode!')
      return
    self._currentBlock.lastRecord()

  def nextRecord(self,block=None):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot go to next record: in query mode!')
      return
    if block:
      block.nextRecord()
    else:
      self._currentBlock.nextRecord()

  def jumpRecord(self,count,block=None):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot jump to record: in query mode!')
      return
    if block:
      block.jumpRecord(count)
    else:
      self._currentBlock.jumpRecord(count)

  def toggleInsertMode(self):
    self._insertMode = not self._insertMode

  def dumpXML(self, treeDump=1, gap="  "):
    return GObj.dumpXML(self, GFParser.getXMLelements(), treeDump, gap)

  def exit(self):
    self._app.executeExit(None)

  def triggerSetFocus(self,object):
    self.changeFocus(object._object,0)

  def triggerSetStatusText(self,tip=''):
    self.dispatchEvent(GFEvent('uiUPDATESTATUS',[tip,None,None,None,None,None,None]))

  def triggerGenericBox(self,message, buttonSet):
    event = GFEvent('genericBox', [message,buttonSet])
    self.dispatchEvent(event)
    return event.data[0]

  def triggerCallTrigger(self, triggerName):
    try:
      self._triggerDictionary[triggerName](self)
    except:
      print "cannot find %s in %s" % (triggerName, self._triggerDictionary.keys ())
      raise

