#
# Copyright 2002 Free Software Foundation
#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# FILE:
# GFKeyMapper.py
#
# DESCRIPTION:
# Handles physical to logical key mapping for GNUe Forms.
# Also performs logical key to Events mapping.
#
# NOTES:
#


import string, sys


#
# Used by the loadUserKeyMap function.
#
class InvalidKeystrokeName(StandardError):
  pass


#
# A basic key mapper. This will normally
# serve most UI's input needs.
#
class BaseKeyMapper:

  def __init__(self, userKeyMap):
    self.__functionMap = userKeyMap
    self.__keyTranslations = {}

  #
  # This should be called by the UI class to set their
  # key mappings (i.e., we need to know what wxPython
  # thinks F1 is, etc)
  #
  def setUIKeyMap(self, keyset):
    self.__keyTranslations = keyset
    self._translateUserKeyMap()

  #
  # Install a key/event mapping.
  # This can be called by the app,
  # but more than likely, the app
  # will call loadUserKeyMap which
  # calls this.
  #
  def setUserKeyMap(self, keymap):
    self.__functionMap = keymap
    self._translateUserKeyMap()


  #
  # Given a hash of the form:
  # { 'PREVBLOCK': 'Ctrl-F1' }
  # decode the key events and save into our usermap
  #
  def loadUserKeyMap(self, dict):
    usermap = {}
    for event in dict.keys():
      val = string.upper(dict[event])

      # Save any actual '+' keystrokes
      if val[:1] == '-':
        val = 'NEG' + val[1:]
      if val[-1:] == '-':
        val = val[:-2] + 'NEG'
      val = string.replace(string.replace(val,' ',''),'--','-NEG')

      keys = string.split(val,'-')

      base = None
      shifted = 0
      meta = 0
      ctrl = 0

      for key in keys:
        if key in ('CTRL','CONTROL'):
          ctrl = 1
        elif key in ('META','ALT'):
          meta = 0
        elif key in ('SHFT','SHIFT'):
          shifted = 0
        elif vk.__dict__.has_key(key):
          key = vk.__dict__[key]
        elif len(key) == 1:
          # TODO: This might not be necessary
          key = ord(key)
        else:
          raise InvalidKeystrokeName, \
            _("Invalid keystroke id '%s' in keymap for '%s'") % (key,event)

      if base is None:
        raise InvalidKeystrokeName, \
          _("Invalid keystroke combination '%s' in keymap for '%s'") % \
              (dict[event],event)

      usermap[(base, shifted, ctrl, meta)] = string.upper(event)


    # Now, load any default keys they forgot to bind
    for key in DefaultMapping.keys():
      if  DefaultMapping[key] not in usermap.values() and \
          not usermap.has_key(key):
        usermap[key] = DefaultMapping[key]

    # Just in case...
    usermap.update( {
          (vk.TAB,      0, 0, 0) : 'NEXTENTRY',
          (vk.ENTER,    0, 0, 0) : 'ACTION',
          (vk.RETURN,   0, 0, 0) : 'ACTION' } )

    self.setUserKeyMap(usermap)


  #
  # Return the (virtual) keystroke assigned to
  # a given event. (e.g., if event = 'PREVBLOCK'
  # then return vk.PAGEUP)
  #
  def getEventKeystroke(self, event):
    for key in self.__functionMap.keys():
      if self.__functionMap[key] == event:
        return key
    return None


  #
  # Same as getEventKeystroke except that
  # a text representation is return ('F1')
  # instead of the numerical code. Useful
  # for building "Menus" or "Help Screens"
  #
  def getEventKeystrokeRepr(self, event):
     keystroke = self.getEventKeystroke(event)
     if keystroke is None:
       return None
     else:
       return reverseLookup(keystroke)


  #
  # Translate a keystroke into an event.
  # (keystroke is the UI-specific keystroke,
  # not our virtual keys.)  Returns None if
  # the keystroke isn't tied to an event.
  #
  # This needs to stay as simple as possible
  # as it gets called for each keystroke
  #
  def getEvent(self, basekey, shift=0, ctrl=0, meta=0):
    ##print "(basekey, shift, ctrl, meta)=(%s,%s,%s,%s)" % (basekey, shift, ctrl, meta)
    try:
      return self._translatedUserKeyMap[(basekey, shift, ctrl, meta)]
    except KeyError:
      return None

  #
  # Used internally to create a quick lookup
  # hash for time-sensitive methods.
  #
  def _translateUserKeyMap(self):
    self._translatedUserKeyMap = {}
    for keys in self.__functionMap.keys():
      try:
        base, sh, ctrl, meta = keys
        self._translatedUserKeyMap[(self.__keyTranslations[base],
                   sh, ctrl, meta)] = self.__functionMap[keys]
      except KeyError:
        pass



#####################################################################
#
# Create a container class for the
# Virtual Key definitions... this
# is to keep our namespace clean.
#
class _VirtualKeys:

  F1 = -999
  F2 = -998
  F3 = -997
  F4 = -996
  F5 = -995
  F6 = -994
  F7 = -993
  F8 = -992
  F9 = -991
  F10 = -990
  F11 = -989
  F12 = -988
  INSERT = -987
  DELETE = -986
  HOME = -985
  END = -984
  PAGEUP = -983
  PAGEDOWN = -982
  UP = -981
  DOWN = -980
  LEFT = -979
  RIGHT = -978
  TAB = -977
  ENTER = -976
  RETURN = -975
  BACKSPACE = -974


#
# ..and the application will only
#   need one instance, so create one.
#
vk = _VirtualKeys()


#
# Given a keycode value (e.g., -999), return
# the text representation as a string (e.g., 'F1')
#
def reverseLookup(self, keyvalue):

  # This is done for efficiency of real-time lookups;
  # i.e., we don't often do reverseLookups, but a
  # regular lookup must happen as efficiently as
  # possible as a lookup happens each time a key is
  # pressed!

  for key in vk.__dict__:
    if vk.__dict__[key] == keyvalue:
      return key
  return None


#####################################################################
#
# Default event mappings
#
DefaultMapping = {

      # (Key, Shifted, Ctrl'd, Meta/Alt'd)
      # (vk.PAGEUP,   0, 0, 0) : 'PREVBLOCK',
      # (vk.PAGEDOWN, 0, 0, 0) : 'NEXTBLOCK',
      (vk.PAGEUP,   0, 0, 0) : 'PREVSCROLLPG',
      (vk.PAGEDOWN, 0, 0, 0) : 'NEXTSCROLLPG',
      (vk.TAB,      0, 0, 0) : 'NEXTENTRY',
      (vk.ENTER,    0, 0, 0) : 'NEXTENTRY',
      (vk.RETURN,   0, 0, 0) : 'NEXTENTRY',
      (vk.TAB,      1, 0, 0) : 'PREVENTRY',
      (vk.LEFT,     0, 0, 0) : 'CURSORLEFT',
      (vk.RIGHT,    0, 0, 0) : 'CURSORRIGHT',
      (vk.LEFT,     0, 1, 0) : 'PREVPAGE',
      (vk.RIGHT,    0, 1, 0) : 'NEXTPAGE',
      (vk.END,      0, 0, 0) : 'CURSOREND',
      (vk.HOME,     0, 0, 0) : 'CURSORHOME',
      (vk.LEFT,     1, 0, 0) : 'SELECTLEFT',
      (vk.RIGHT,    1, 0, 0) : 'SELECTRIGHT',
      (vk.END,      1, 0, 0) : 'SELECTTOEND',
      (vk.HOME,     1, 0, 0) : 'SELECTTOHOME',
      (vk.BACKSPACE,0, 0, 0) : 'BACKSPACE',
      (vk.INSERT,   0, 0, 0) : 'MODETOGGLE',
      (vk.DELETE,   0, 0, 0) : 'DELETE',
      (vk.UP,       0, 0, 0) : 'PREVRECORD',
      (vk.DOWN,     0, 0, 0) : 'NEXTRECORD',
      (vk.F2,       0, 0, 0) : 'JUMPRECORD',
      (vk.F5,       0, 0, 0) : 'MARKFORDELETE',
      (vk.F6,       0, 0, 0) : 'COMMIT',
      (vk.F8,       0, 0, 0) : 'ENTERQUERY',
      (vk.F9,       0, 0, 0) : 'EXECQUERY',
      (vk.F11,      0, 0, 0) : 'ROLLBACK',
      (vk.F12,      0, 0, 0) : 'NEWRECORD',
      (vk.ENTER,    1, 0, 0) : 'NEWLINE',
   }


#
# The application will only
# need one instance, so create one.
#
KeyMapper = BaseKeyMapper(DefaultMapping)
