#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GFEntry.py
#
# DESCRIPTION:
# The primary data entry widget in forms
#
# NOTES:
#

from gnue.common import GDebug, FormatMasks
from gnue.forms import GFDisplayHandler
from GFValue import GFValue
import string

############################################################
# GFEntry
#
# Matches the Entry form tag in the gfd
#
# It send events to its parent GFBlock
#
class GFEntry(GFValue):
  def __init__(self, parent=None, value=None):
    GFValue.__init__(self, parent, value, 'GFEntry')

    # Default attributes (these may be replaced by parser)
    self.height = float(gConfig('widgetHeight'))
    self.width = float(gConfig('widgetWidth'))
    self.typecast="text"
    self.case="mixed"
    self.style="default"

    # Runtime variables
    self._uppercase = 0
    self._lowercase = 0
    self._numeric = 0
    self._queryable = 1

    self._inits = [self.initialize]
    self._allowedValues = None
    self._allowedValuesDescr = [""]

    self._maskHandler = None

    self._rows = 1
    self._gap = 0

    #
    # Trigger exposure
    #
    self._triggerFunctions={'allowedValues':{'function':self.allowedValues,},
                            'atomicModify':{'function':self.triggerAtomicModify},
                            'postAtomicModify':{'function':self.triggerPostAtomicModify},
                            }
    self._triggerProperties={'readonly':{'set':self.setReadonly,
                                         'get':self.getReadonly
                                         },
                             }
    self._triggerSet = self.triggerSetValue
    self._triggerGet = self.getValue


  def triggerAutofillBySequence(self,sequenceName):
    if (not self.getValue()) or self.getValue()=="":
      sequenceNumber = self._block._dataSourceLink._dataObject.triggerExtensions.getSequence(sequenceName)
      self.setValue(sequenceNumber)

  def triggerPostAtomicModify(self,func):
    self.triggerAtomicModify(func,0)

  def triggerAtomicModify(self,func, autoCommit=1):
    if self.getValue() or self.getValue()!="":
      self._block._dataSourceLink._dataObject.triggerExtensions.atomicModifyForEntry(func=func, \
                                                                           field=self.field, \
                                                                           dataObject=self._block._dataSourceLink._dataObject,\
                                                                           record=self._block._resultSet.current, \
                                                                           cond = None, \
                                                                           autoCommit = autoCommit)

  def getReadonly(self):
    return self.readonly

  def setReadonly(self,value):
    self.readonly = value

  def _buildObject(self):
    # Convert deprecated attributes

    # Numeric attributes (now typecast="..")
    if hasattr(self,'numeric') and self.numeric:
      del self.numeric
      self.typecast = 'number'

    # visibleCount attributes... (now rows="..")
    if not hasattr(self, 'rows') and hasattr(self,'visibleCount'):
      self.rows = self.visibleCount
      del self.visibleCount

    # Case attributes... (now case="..")
    if hasattr(self,'uppercase') and self.uppercase:
      del self.uppercase
      self.case = 'upper'
    elif hasattr(self,'lowercase') and self.lowercase:
      del self.lowercase
      self.case = 'lower'

    # Force 'label' styles into "readonly" mode
    if self.style == 'label':
      self.readonly = 1
      self._queryable = 0

    return GFValue._buildObject(self)

  def initialize(self):
    if self.typecast == 'number':
      self._numeric = 1
      self._allowFormulas = gConfig("AllowNumericFormulas")
    if self.case == 'upper':
      self._uppercase = 1
    elif self.case == 'lower':
      self._lowercase = 1

    self._formatmask = ""
    self._inputmask = ""
    if hasattr (self, 'displaymask'):
      self._displaymask= self.displaymask
      del (self.displaymask)
    else:
      self._displaymask = ""
    #self._displaymask = ""

    self._block = self.findParentOfType('GFBlock')

    if hasattr(self, 'default') and self.default != None and len(self.default):
      default = self.default
    else:
      default = None

    if not hasattr(self,'field') or not len(self.field):
      self.field = "__GNUe__%s" % self.name
      self._block._dataSourceLink.referenceUnboundField(self.field, default)
      self._bound = 0
    else:
      self._block._dataSourceLink.referenceField(self.field, default)
      self._bound = 1

    if hasattr(self,'foreign_key'):
      foreignKeyContent = string.split(self.foreign_key,'.')
      if len(foreignKeyContent) > 2:
        self.datasource = \
           self._block._form._datasourceDictionary[string.lower(foreignKeyContent[0])]
        #self.tableName = foreignKeyContent[1]
        self.fieldName = foreignKeyContent[1]+'.'+foreignKeyContent[2]
      else:
        self.datasource = \
           self._block._form._datasourceDictionary[string.lower(foreignKeyContent[0])]
        self.fieldName = foreignKeyContent[1]

    if hasattr(self, 'queryDefault') and \
         self.queryDefault != None and \
         self._bound and \
         len(self.queryDefault):
      self._block._queryDefaults[self] = self.queryDefault


    # Create an appropriate display handler
    if self.style == 'checkbox':
      self._displayHandler = GFDisplayHandler.CheckboxDisplayHandler(self,
               self._block._form._app.eventController, self._maskHandler)
    elif self.style == 'dropdown':
      self._displayHandler = GFDisplayHandler.DropdownDisplayHandler(self,
               self._block._form._app.eventController, self._maskHandler)
    elif self.style == 'textlookup':
      self._displayHandler = GFDisplayHandler.TextLookupDisplayHandler(self,
               self._block._form._app.eventController, self._maskHandler)
    elif self.typecast == 'text':
      self._displayHandler = GFDisplayHandler.TextDisplayHandler(self,
               self._block._form._app.eventController,
               self._inputmask, self._displaymask, self._formatmask)
    elif self.typecast == 'number':
      self._displayHandler = GFDisplayHandler.NumberDisplayHandler(self,
               self._block._form._app.eventController,
               self._displaymask, self._inputmask)
    elif self.typecast == 'date':
      self._displayHandler = GFDisplayHandler.DateDisplayHandler(self,
               self._block._form._app.eventController, self._inputmask, self._displaymask)

  def isQueryable(self):
    return self._queryable


  #
  # Pulls the proper value from the resultSet
  #
  def getValue(self):
    # TODO: No longer true...
    # We have to check to see if the _block has been
    # setup yet.  The Trigger system calls getValue
    # before it's ready so we fake it
    #if hasattr(self,'_block'):
    if self._block.mode == 'query':
          try:
            value = self._block._queryValues[self]
          except KeyError:
            value = None
    elif self._block.mode == 'precommit':
        value = self._block._preCommitWorkingRecord.getField(self.field)

    else:
      try:
        value = self._block._resultSet.current.getField(self.field)
        if value == None:
          if hasattr(self._block._resultSet._dataObject,'table'):
            value = self._block._resultSet.current.getField(self._block._resultSet._dataObject.table+'.'+self.field)
      except:
        print self.name
        print self._block.name
        # print self.field.name
        raise
    # TODO: part of the removed block check from above
    #else:
    #  value = None

    if value == None:       value = ''

    # TODO: check to see if this is still needed
    #CheckBox does not work if comment following "if" block (ra3vat)
    if self.style == 'checkbox' and value == '':      value = 0

    return value

  #
  # triggerSetValue
  #
  # does proper typecasting when value is set via a trigger
  #
  def triggerSetValue(self, value):
    if self.style=='checkbox' and value =='':
      value = 0
    if self.typecast == "number" and value != None:
      try:
        value = float(value)
      except ValueError:
        value = ''
    self.setValue(value)

  def setValue(self, value):
    # TODO: This is now being done in the display handler
    #if self.style=='checkbox' and value =='':
    #  value = 0
    #
    # Number fields
    #if self.typecast == "number" and value != None:
    #  value = float(value)

    if self._block.mode == 'query':
      self._block._queryValues[self] = value

    elif self._block.mode == 'precommit':
      value = self._block._preCommitWorkingRecord.setField(self.field,value)

    else:
      self._block._resultSet.current.setField(self.field,value)

#     if self.name=="payEntry":
#       print 'entry::setValue: >%s<' % value
#       from traceback import print_stack
#       import sys
#       print_stack (file=sys.stdout)
#       print '-------------'

    self._displayHandler.setValue (value)

    self._block._form.updateUIEntry(self)

  #
  # allowedValues
  #
  def allowedValues(self):
    if not self._allowedValues and hasattr(self,'datasource'):
      rs = self.datasource.createResultSet()
      more = rs.firstRecord()
      self._allowedValues = {"":""}
      self._allowedValuesDescr = [""]
      self._allowedValuesReverse = {"":""}
      while more:
        try:
          if hasattr(self,'foreign_key_description'):
            key = "%s" % rs.current.getField(self.fieldName)
            descr = "%s" % rs.current.getField(self.foreign_key_description)

            self._allowedValues[key] = descr
            self._allowedValuesDescr.append(descr)
            self._allowedValuesReverse[descr] = key
          else:
            # No descriptions defined
            self._allowedValues["%s" % rs.current.getField(self.fieldName)] = ""
        except AttributeError:
          self._allowedValues[rs.current.getField(self.fieldName)] = None
        more = rs.nextRecord()
##      self._allowedValuesDescr.sort()
      GDebug.printMesg (5,'Created for DropDown: %s' % self._allowedValues)

    return (self._allowedValues, self._allowedValuesDescr)
