#
# Copyright 2000-2002 Free Software Foundation
#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# FILE:
# UIbase.py
#
# DESCRIPTION:
# A base class for UI modules.  Everything in here better
# not contain any specific UI widgetisms
#
# NOTES:
#

from gnue.forms.GFForm import *
from gnue.forms.GFEvent import *
from gnue.common.GObjects import *

# mdione
import sys

# Used in updating the status bar

statusValues = {'saved':'OK',
                'modified':'MOD',
                'deleted':'DEL',
                'query':'QRY'}

#
# GFUserInterfaceBase
#
# Convience class to be inheritance by other GFUserInterface's
# Provides basic functionality.
#
class GFUserInterfaceBase(GFEventAware):
  def __init__(self, eventHandler):
    GFEventAware.__init__(self, eventHandler)
    #
    # Events I monitor...
    #
    self.registerEventListeners( {
            'gotoENTRY'        : self.switchFocus,
            'updateENTRY'      : self.updateEntry,
            'updateEntryEditor': self.updateEntryEditor,
            'gotoPAGE'         : self.gotoPage,
            'formALERT'        : self.formAlert,
            'beginWAIT'        : self.beginWait,
            'endWAIT'          : self.endWait,
            'uiUPDATESTATUS'   : self.updateStatusBar,
            'uiABOUT'          : self.aboutBox,
            'exitApplication'  : self.exitApplication,
            'msgBoxUI'         : self.msgBox,
            'genericBox'       : self.genericBox,
            # Clipboard contents
            'setCLIPBOARD'     : self.setClipboardContents,
            'getCLIPBOARD'     : self.getClipboardContents,
          })

    # Our local "clipboard"
    self.__clipboard = None

    self.widgets = self._WIDGETS

    self._formToUI = {}                   # the GFObj to UI widget cross ref
    self._formToUIWidget = {}             # dictionary of driver specific UIfoo
                                          # widgets with the GFObj used as the key

    self.containerWidgets= ['GFPage', 'GFBox']     # GFForm objects that contain others

    self._pageList = []                   # A list of page widgets created by the form

    self.keyEvents = {}                   # Keyboard events from the widget set the
                                          # UI is expected to respond to

    # A dictionary of functions used by buildUI to build the widgets.  Can override in
    # UI specific module if necessary.
    self.widgetConstructorFunction = {
      'GFBox'         : self._stdHandler,
      'GFButton'      : self._stdHandler,
      'GFEntry'       : self._stdHandler,
      'GFForm'        : self._stdHandler,
      'GFLabel'       : self._stdHandler,
      'GFScrollBar'   : self._stdHandler,
      'GFPage'        : self._stdHandler,
      'GFMainToolBar' : self._stdHandler,
      'GFMainMenuBar' : self._stdHandler,
      }

  def _stdHandler(self):
    object = self._currentObject[0]
    event = self._updateCreateEvent(GFEvent('CreateUIWidget',None,
                   object=object,
                   container=self.currentWidget[0],
                   textWidth=self.textWidth,
                   textHeight=self.textHeight,
                   widgetWidth=self.widgetWidth,
                   widgetHeight=self.widgetHeight,
                   interface=self,
                   initialize=1))

    newWidget = self.widgets[object._type](event)
    return newWidget

  # Can be used by UI drivers to add more attributes to the event.
  def _updateCreateEvent(self, event):
    return event


  #
  # buildUI
  #
  # Creates the user interface from the form definition
  # passed to the GFUserInterface instance.  Creates a
  # form to UI cross reference used by the UI to match
  # form objects to the UI counterparts
  #
  def _buildUI(self):
    try:
      uiWidget = self.widgetConstructorFunction[self._currentObject[0]._type] ()
      widget = uiWidget.widgets
      GDebug.printMesg(2, "Widget is %s" % widget)
      self._formToUIWidget[self._currentObject[0]]=uiWidget

      # If the widget list is not empty then do the following
      # list would be empty on hidden fields
      if not widget == []:
        # Create a cross reference to the GFObject that created
        # one or more UI widgets
        self._formToUI[self._currentObject[0]] = widget
        # If the current object is a container then add the
        # just created first created widget to it's widget list
        if self._currentObject[0]._type in self.containerWidgets:
          self.currentWidget.insert(0,widget[0])

    except KeyError:
      # mdione
      GDebug.printMesg (2, "_buildUI failed on a %s: %s" % (self._currentObject[0]._type, sys.exc_info()[0]))
      pass


    if self._currentObject[0]._children:
     for child in self._currentObject[0]._children:
        if isinstance(child, GObj):
          self._currentObject.insert(0,child)
          self._buildUI()
          self._currentObject.pop(0)

    if self._currentObject[0]._type in self.containerWidgets:
      self.currentWidget.pop(0)

  #############################################################################
  #
  # Public Interface
  #
  # The interface exposed to the forms backend
  #
  #

  #
  # buildForm
  #
  # when passed a GObj tree constructed from a .gfd file
  # it handles the creatation of the user interface
  #
  def buildForm(self, form):
    self._currentObject = [form]
    self._form = form
    # Create the UI from the GFForm passed in
    self._buildUI()




  #############################################################################
  #
  # Generic Incoming Event Handlers
  #
  # Handles incoming events and calls UI instance specific functions to
  # execute the actions.  These events come from the forms back end.
  #

  #
  # Clipboard routines
  #
  # If a particular UI has a system-wide clipboard,
  # these methods should be overridden to use that
  # clipboard.
  #
  def getClipboardContents(self, event):
    GDebug.printMesg(5, "Getting clipboard '%s'" % self.__clipboard)
    event.__result__ = "%s" % self.__clipboard

  def setClipboardContents(self, event):
    GDebug.printMesg(5, "Setting clipboard '%s'" % event.text)
    self.__clipboard = "%s" % event.text

  #
  # exitApplication
  #
  # Form has told the application to close so call the UIs private exit routine
  #
  def exitApplication(self,event):
    self._exit()

  #
  # updateStatusBar
  #
  # Calls the UIs function to update status
  def updateStatusBar(self,event):

    tip, recordStatus, insertMode, currentRecord, maxRecord, currentPage, maxPage =event.data

    status = None
    if recordStatus != None:
      status = statusValues[recordStatus]

    insertValue = None
    if insertMode != None:
      insertValue = insertMode and 'INS' or  'OVR'
    self._setStatusBar(tip, status, insertValue, currentRecord, maxRecord, currentPage, maxPage)

  #
  # exitApplication
  #
  # Forces closure of the form by passing a non-vetoable close
  # to the application's frame.  closeTrap will then exit w/o
  # asking the form for permission again.
  #
  def exitApplication(self,event):
    self._exit()

  #
  # switchFocus
  #
  # Moves the focus to a specific UI widget
  #
  def switchFocus(self, event):
    formObject = event.data['object']
    if formObject: # Some pages might not have any widgets that can be active
      index = formObject._visibleIndex
      self._formToUIWidget[formObject].indexedFocus(index)

    self.dispatchEvent(GFEvent('beginEDITMODE', formObject))


  #
  # aboutBox
  #
  # Displays a modal about box on the screen.
  #
  def aboutBox(self, event):
    programVersion = event.data[0]
    formName = event.data[1]
    formVersion = event.data[2]
    author = event.data[3]
    description = event.data[4]

    about = self._DIALOGS['about'](self.mainWindow, programVersion, formName, formVersion, author, description)
    about.showModal()
    about.destroy()

  #
  # msgBox
  #
  def msgBox(self,event):
    message = event.data.message
    caption = event.data.caption
    
    dialog = self._DIALOGS['messageBox'](self.mainWindow, message, caption)
    dialog.showModal()
    dialog.destroy()


  #
  # genericBox
  #
  def genericBox(self,event):
      message = event.data[0]
      caption = "Aviso"
      buttonSet = event.data[1]
      if len(event.data) > 2:
        defaultButton = event.data[2]
      else:
        defaultButton = -1
      if len(event.data) > 3:
        cancelButton = event.data[3]
      else:
        cancelButton = -1

      self.genericBox = uiDlg = self._DIALOGS['genericBox'](self.mainWindow, caption)
      self.genericBox = dlg = uiDlg.createGenericBox(message, buttonSet)
      
      dlg.ShowModal()
      
      event.data = [uiDlg.getResult()]

  #
  # updateEntryEditor
  #
  def updateEntryEditor(self, event):
    index = event.object._visibleIndex
    widget = self._formToUIWidget[event.object]
    widget.setValue(event.display, index)
    widget.setCursorPosition(event.cursor, index)

    if event.selection != None:
      selection1, selection2 = event.selection
      widget.setSelectedArea(selection1, selection2, index)


  #
  # updateEntry
  #
  def updateEntry(self,event):
    if event.data.hidden or event.data._type == 'GFButton':
      return

    entry = event.data
    handler = entry._displayHandler
    prehandling = handler.editing
    handler.editing = 0
    index = entry._visibleIndex
    block = entry._block
    currentRecord = block._resultSet.getRecordNumber()

    # Fill the prior spots
    for count in range(index):
        value = handler.getDisplayFiller(block._resultSet.getRecord(currentRecord- \
                                (index-count)).getField(entry.field))

        self._formToUIWidget[entry].setValue(value,count)


    # Fill current spot
    value = handler.getDisplayFiller(entry.getValue())
    self._formToUIWidget[entry].setValue(value, index)

    # Fill trailing spots
    #
    # You must skip the matching index but
    # you do not want to just add 1 to count
    # as the formulas would then be off
    count = index
    while count < int(entry._rows):
      if count != index:
        rec = block._resultSet.getRecord(currentRecord+(count-index))

        if rec == None:
          value = handler.getDisplayFiller(None)
        else:
          value = handler.getDisplayFiller(rec.getField(entry.field))

        self._formToUIWidget[entry].setValue(value, count)
      count += 1

    handler.editing = prehandling


  # Called whenever forms goes into a "wait" state in which user cannot
  # interact with interface (e.g., while waiting for a query or a commit)
  def beginWait (self, event):
    pass


  # Called whenever forms leaves a "wait" state
  def endWait (self, event):
    pass






