#
# Copyright 2000-2002 Free Software Foundation
#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# FILE:
# UIdriver.py
#
# DESCRIPTION:
# A nstti based user interface driver for GNUE forms.
# (nstti : The Not So Tiny Text Interface)
# http://tomte.freezope.org/td_content_level_1/code_snippets/td_content_level_2/0_nstti/
#
# NOTES:
#

from nstti import *
import curses

from GFForm import *
from GFEvent import *
from gnue.forms.uidrivers._base.UIdriver import *
from gnue.forms.uidrivers._base.UIWidget import *

import GFKeyMapper

import os
import sys
import string
from gnue.forms import VERSION
from gnue.common import GDebug, GLoginHandler
from gnue.common import GConfig
from gnue.common import lineWrap

_NOTEBOOK = None
_LOOPTRAP = 0
_EVENTPROCESSOR = None
_charWidth = 0
_IdToWxObj = {}
_IdToGFObj = {}
_IdToUIObj = {}

#
# GFUserInterface
#
# The public interface to the User Interface
# All UIs must provide this class
#
class GFUserInterface(GFUserInterfaceBase):
  def __init__(self, eventController, disableSplash = None):
    self._WIDGETS = {'GFLabel'     : UILabel,
                     'GFBox'       : UIBox,
                     'GFPage'      : UIPage,
                     'GFForm'      : UIForm,
                     'GFEntry'     : UIEntry,
                     'GFButton'    : UIButton,
                     'GFScrollBar' : UIScrollBar,
                     }
    
    self._DIALOGS = {'about'       :UIAbout,
                     'messageBox'  :UIMessageBox,
                     }

    self._aboutDialog = UIAbout
    self._messageDialog = UIMessageBox

    self._disableSplash = disableSplash
    GFUserInterfaceBase.__init__(self,eventController)

    global _EVENTPROCESSOR
    _EVENTPROCESSOR = self.dispatchEvent

    self.forms = {}
    self.init()

  #
  # init
  #
  # Routine called by wxWindows when wxApp instance is created
  # It is used the build the basic UI
  #
  def init(self):
    self._init()
    
  def _init(self):
    #
    # SplashScreen
    #
    if not self._disableSplash:
      text = "GNU Enterprise Forms"
      text = "Version %s" % VERSION
      text = "(c)2000-2002 Free Software Foundation"
      text = "Protected by GNU General Public Licence 2.0"

    #
    # Create the main window
    #
    self._screen = sio()
    self._screen.BeginScreenAccess()
    self.mainWindow = ScreenContainer(self._screen)
    self._screen.Cls()
    self.mainWindow.panel = Dialog(self.mainWindow, 0, 0, self._screen.MAXROW, self._screen.MAXCOL )

    self.currentWidget = [self.mainWindow.panel]

     
    menu = {'file':[ {'id'   :100,
                      'text' :'Commit     (F6)',
                      'tip'  : "gConfig('msg_commit')",
                      'event':'requestCOMMIT'},
                     {'id'   :102,
                      'text' :'Rollback  (F11)',
                      'tip'  : "gConfig('msg_rollback')",
                      'event':'requestROLLBACK'},
                     {'id': -1,'text': None, 'tip': None, 'event':None},
                     {'id'   :104,
                      'text' :'Print          ',
                      'tip'  : "gConfig('msg_print')",
                      'event':'requestPRINTOUT'},
                     {'id'   :101,
                      'text' :'Exit           ',
                      'tip'  : "gConfig('msg_exit')",
                      'event':"self.menuExitEvent"},
                     ],
            'edit':[ {'id'   :401,
                      'text' :'Cut',
                      'tip'  :'Cut the selected text',
                      'event':'requestCUT'},
                     {'id'   :402,
                      'text' :'Copy',
                      'tip'  : "Copy the selected text",
                      'event':'requestCOPY'},
                     {'id'   :403,
                      'text' :'Paste',
                      'tip'  : "Paste the selected text",
                      'event':'requestPASTE'},
                     ],
            'data':[ {'id'   :201,
                      'text' :'Next Record',
                      'tip'  :"gConfig('msg_next')",
                      'event':'requestNEXTRECORD'},
                     {'id'   :202,
                      'text' :'Previous Record',
                      'tip'  :"gConfig('msg_previous')",
                      'event':'requestPREVRECORD'},
                     {'id'   :209,
                      'text' :'Jump to Record...',
                      'tip'  :"gConfig('msg_jump')",
                      'event':'self.promptForRecordNumber'},
                     {'id': -1,'text': None, 'tip': None, 'event':None},
                     {'id'   :204,
                      'text' :'Delete Record          (F5)',
                      'tip'  :"gConfig('msg_delete')",
                      'event':'requestMARKFORDELETE'},
                     {'id'   :205,
                      'text' :'Next Block           (PgDn)',
                      'tip'  :"gConfig('msg_next_block')",
                      'event':'requestNEXTBLOCK'},
                     {'id'   :206,
                      'text' :'Previous Block       (PgUp)',
                      'tip'  :"gConfig('msg_previous_block')",
                      'event':'requestPREVBLOCK'},
                     {'id': -1,'text': None, 'tip': None, 'event':None},
                     {'id'   :207,
                      'text' :'Enter Query            (F8)',
                      'tip'  :"gConfig('msg_query_prep')",
                      'event':'requestENTERQUERY'},
                      {'id'   :208,
                      'text' :'Execute Query        (F9)',
                      'tip'  :"gConfig('msg_query')",
                      'event':'requestEXECQUERY'},
                     ],
            'help':[ {'id'   :301,
                      'text' :'About...',
                      'tip'  :"gConfig('msg_help')",
                      'event':'requestAbout'},
                     ],
            }

    self.widgetWidth = 1  # The width of a 1 char widget (for things like buttons)
    self.widgetHeight = 1 # The pixel height of a 1 char widget (for things like buttons)
    self.textWidth = 1    # The pixel width of text inside a widget
    self.textHeight = 1   # The pixel height of text inside a widget
            

    menuBar = MenuBar(self.mainWindow,0,0)

    for _menu in menu.keys():
      newMenu=menuBar.NewMenu(_menu,_menu)
      for menuitem in menu[_menu]:
        if menuitem['id'] > 0:
          newMenu.AddItem(menuitem['id'],menuitem['text'])
        else:
          newMenu.InsertSeperator()
      newMenu.Finalize()

    self.mainWindow.AddControl(menuBar)
    return 1

  def activateForm(self, form):
    width = int(self._form.width)
    height = int(self._form.height)
    pass
  
##     formSize = wxSize(width*int(self.widgetWidth),
##                       int(height+self.menu_sb_space)*int(self.widgetHeight))

##     self.mainWindow.SetSize(formSize)
##     self.mainWindow.SetTitle(str(self._form.title))
##     self.mainWindow.panel =  self._pageList[0]
##     self.mainWindow.panel.SetSize(formSize)
##     self._pageList[0].Show(1)

##     # Only one page at a time can be visible
##     self.visiblePage = self._pageList[0]

##     if _NOTEBOOK: # Adjust sizes
##       _NOTEBOOK.SetSize(formSize)

##       child = _NOTEBOOK.GetParent()
##       while child:
##         child.Fit()
##         child = child.GetParent()

##     self.mainWindow.CenterOnScreen()


  #############################################################################
  #
  # Private UIBase support functions
  #
  # Called in UIbase functions to perform UI interaction with this specific
  # widget set.
  #

  #
  # _exit
  #
  # Tells the application to close it's main window
  #
  def _exit(self):
    self.mainWindow.EndScreenAccess()


  def _setStatusBar(self,*args):
    pass
  #############################################################################
  #
  # Incoming Event Processors
  #
  # Processes the incoming events from other objects
  #

  #
  # mainLoop
  #
  # The primary loop of the user interface.  Called once the UI is
  # fully activated
  #
  def mainLoop(self):
    self.mainWindow.Run()

  #
  # gotoPage
  #
  # makes the requested page visible on the screen
  #
  def gotoPage(self, event):
    pass
##     if not _NOTEBOOK:
##       self.visiblePage.Show(FALSE)
##       self.visiblePage = self._formToUI[event.data][0]
##       # TODO: move the calculatoin out of here
##       self.visiblePage.SetSize(wxSize(int(self._form.width)*int(self.widgetWidth),
##                                       int(int(self._form.height)+self.menu_sb_space)*int(self.widgetHeight)))
##     else:
##       self.visiblePage = self._formToUI[event.data][0]
##       for count in range(_NOTEBOOK.GetPageCount()):
##         if self.visiblePage.GetId() == _NOTEBOOK.GetPage(count).GetId():
##           _NOTEBOOK.SetSelection(count)
##           break
##     self.visiblePage.Show(1)

  #
  # formAlert
  #
  # Rings a bell and alters the statusbar to display
  # a line of text
  #
  def formAlert(self, event):
    print "Beep! Beep!"
    
  # Called whenever forms goes into a "wait" state in which user cannot
  # interact with interface (e.g., while waiting for a query or a commit)
  def beginWait (self, event):
    print "Dew Break!"

  # Called whenever forms leaves a "wait" state
  def endWait (self, event): 
    print "Off yer keesteers!"
    

  #############################################################################
  #
  # Internal Event Processors
  #
  # Processes the events from the widget set
  #

  #
  # menuExitEvent
  #
  # Catches the Exit event from the menu and make sure it closes in such a way to
  # trigger the closeTrap method
  #
  def menuExitEvent(self, event):
    self.mainWindow.Close()

  #
  # promptForRecordNumber
  #
  # Presents a dialog to user asking for the record number to jump to
  #
  def promptForRecordNumber(self, event= None):
     pass

    
  #
  # closeTrap
  #
  # intercepts the applications closure and generates an event to the form requesting
  # closure.  Allows the form to control closure.  If the form approves it'll send
  # back an event closing the application
  #`
  def closeTrap(self,event):
    pass
##     object = _eventObjTowxWindow(event)
##     if event.CanVeto():
##       self.dispatchEvent(GFEvent('requestEXIT'))
##     else:
##       object.Destroy()

#############################################################################
#############################################################################
#
# UI Objects
#
# A set of standard UI objects that allow us to move more code into UIbase
# These map generic functionality to the specific UI widget set in use by
# this driver. 
#

#############################################################################
#############################################################################

#
# UIHelper
#
# Provides a UI widget set of std functions to reduce the number of functions
# require in each UIclass listed below
class UIHelper:
  def showModal(self):
    pass # self.ShowModal()

  def show(self, flag=1):
    pass # self.Show(flag)

  def destroy(self):
    pass # self.Destroy()

  def indexedFocus(self, index):
    pass # self.widgets[index].SetFocus()

  def setValue(self, value, index=0):
    pass
##     self.widgets[index].SetEvtHandlerEnabled(FALSE)
##     self.widgets[index].SetValue(value)
##     self.widgets[index].SetEvtHandlerEnabled(1)

  def setCursorPosition(self, position, index=0):
    pass
##     try:
##       self.widgets[index].SetInsertionPoint(position)
##     except AttributeError:
##       pass  # For label-style entries

  def setSelectedArea(self, selection1, selection2, index=0):
    pass
##     try:
##       self.widgets[index].SetSelection(selection1, selection2)
##     except AttributeError:
##       pass  # For label-style entries

  def _addToCrossRef(self, widget,gfobject, uiobject):
    pass
##     id = widget.GetId()
##     _IdToWxObj[id]=widget
##     _IdToGFObj[id]=gfobject
##     _IdToUIObj[id]=uiobject

  def createWidget(self, event, spacer):

    newWidget = self._createWidget(event, spacer)
    self._addToCrossRef(newWidget,event.object,self)
    return newWidget

#
# UILabel
#
# Widget set specific function that creates a single instance of a label
#
class UILabel(UIHelper, UIWidget):
  alignmentStyle = { 'left'  : 0,
                     'center': 0,
                     'right' : 0,
                     }

  def _createWidget(self, event, spacer):
    object = event.object
    style = self.alignmentStyle[object.alignment]
    newWidget = Label(event.container,'name',object.y,object.x,str(object.text))
    return newWidget

#
# UIBox
#
# Widget set specific function that creates a single instance of a border box
#
class UIBox(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object
    newWidget = Frame(event.container,'name',object.y,object.x,
                      object.width, object.height,str(object.label))
    return newWidget

#
# UIScrollBar
#
# Widget set specific function that creates a single instance of a scroll bar used
# to navigate multiple records
#
class UIScrollBar(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object

    if object.width > 1:
      newWidget = ScrollBar(event.container,'name',object.y,object.x,object.width)
    else:
      newWidget = VScrollBar(event.container,'name',object.y,object.x,object.height)
    return newWidget

#
# UIButton
#
# Widget set specific function that creates a single instance of a button
#
class UIButton(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object
    newWidget = Button(event.container,'name',object.y,object.x,object.width,str(object.label))
    return newWidget

##   def buttonHandler(self,event):
##     if event.GetEventType() == wxEVT_COMMAND_BUTTON_CLICKED:
##       action = None
##       object = _eventObjTowxWindow(event)

##       id = object.GetId()
##       gfObject     = _IdToGFObj[id]
##       action = GFEvent('fireTRIGGER',gfObject.trigger)

##       if action:
##         _EVENTPROCESSOR(action)

#
# UIPage
#
# Widget set specific function that creates a single instance of a page
#
# Note: The event trap for changing notebook pages is attached to the UIForm
#
class UIPage(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object
    newWidget = Dialog(container, 0, 0, event.widgetWidth, event.widgetHeight )
    newWidget.DeActivate(tiFALSE)

    event.interface._pageList.append(newWidget)
    return newWidget

#
# UIEntry
#
# Widget set specific function that creates a single instance of a data entry widget
#
class UIEntry(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object

    style = object.style
    if style == 'dropdown':
      if event.initialize:
        choices = object.allowedValues()[1]
      else:
        choices = [""]
      newWidget = ComboBox(event.container, 'name',object.y,object.x, object.width)
      for index in range(len(choices)):
        newWidget.AddItem(index,choices[index])
    elif style == 'label':
      newWidget = TextBox(event.ontainer, 'name', object.y, object.x, object.width)
    elif style == 'checkbox':
      newWidget = CheckBox(event.container, 'name',object.y,object.x,'',0)
    else: # Normal text box
      newWidget = TextBox(event.container, 'name', object.y, object.x, object.width)
    return newWidget

#
# UIForm
#
# Widget set specific function that creates a single instance of a Form widget
#
class UIForm(UIHelper, UIWidget):
  tabStyles = {'top':0,
               'left':1,
               'right':2,
               'bottom':3,
               }

  def _createWidget(self, event, spacer):
    object = event.object
    newWidget = Dialog(event.container, 0, 0, event.interface._screen.MAXROW, event.interface._screen.MAXCOL )

    return newWidget


#####################################################################
##
## Required Dialog box support
##
#####################################################################

#
# UIAbout
#
# A dialog box providing about info
#
class UIAbout( UIHelper):
  def __init__(self, interface, programVersion, formName, formVersion, author, description):
    pass

# UImessageBox
#
# A dialog box providing basic message info
#
class UIMessageBox(UIHelper):
  def __init__(self, interface, message, caption):
    message = lineWrap(message,60)
    pass
    
#####################################################################
##
## Keymapper Support
##
#####################################################################
from GFKeyMapper import vk

# Translate from curses keystrokes to our virtual keystrokes
cursesKeyTranslations = {
  vk.F1        : curses.KEY_F1,        vk.F2       : curses.KEY_F2,
  vk.F3        : curses.KEY_F3,        vk.F4       : curses.KEY_F4,
  vk.F5        : curses.KEY_F5,        vk.F6       : curses.KEY_F6,
  vk.F7        : curses.KEY_F7,        vk.F8       : curses.KEY_F8,
  vk.F9        : curses.KEY_F9,        vk.F10      : curses.KEY_F10,
  vk.F11       : curses.KEY_F11,       vk.F12      : curses.KEY_F12,
  vk.INSERT    : curses.KEY_IC,        vk.DELETE   : curses.KEY_DC,
  vk.HOME      : curses.KEY_HOME,      vk.END      : curses.KEY_END,
  vk.PAGEUP    : curses.KEY_NPAGE,     vk.PAGEDOWN : curses.KEY_NPAGE,
  vk.UP        : curses.KEY_UP,        vk.DOWN     : curses.KEY_DOWN,
  vk.LEFT      : curses.KEY_LEFT,      vk.RIGHT    : curses.KEY_RIGHT,
  vk.TAB       : curses.KEY_CTAB,      vk.ENTER    : curses.KEY_ENTER,
  vk.BACKSPACE : curses.KEY_BACKSPACE }


GFKeyMapper.KeyMapper.setUIKeyMap(cursesKeyTranslations)

#####################################################################
##
## Basic Event Processing
##
#####################################################################

def _setDefaultEventHandlers(newWidget, initialize):
  global _EVENTPROCESSOR
##   if initialize:
##     # TODO: this should use one instance
##     newWidget.PushEventHandler(mouseEvtHandler(_EVENTPROCESSOR))
##     newWidget.PushEventHandler(keyboardEvtHandler(_EVENTPROCESSOR))

class uiBaseEvtHandler:
  def __init__(self,eventList,eventProcessor):
    pass
##     wxEvtHandler.__init__(self)
##     self._eventProcessor = eventProcessor
##     for eventType in eventList:
##       self.Connect(-1, -1, eventType,self._processEvent)
                        
##   def _processEvent(self,event):
##     GDebug.printMesg(0,"uiBaseEvtHandler _processEvent was called?!?")
  
class mouseEvtHandler(uiBaseEvtHandler):
  def __init__(self,eventProcessor):
    pass
##     eventList = [wxEVT_LEFT_DOWN]
##     uiBaseEvtHandler.__init__(self, eventList, eventProcessor)
    
##   def _processEvent(self,event):
##     global _charWidth
##     object = _eventObjTowxWindow(event)
    
##     # compute the location of the character in the widget
##     x,y = event.GetPosition()
##     cursorPosition = 1 + x/_charWidth

##     # Move to proper object
##     id = object.GetId()
##     gfObject     = _IdToGFObj[id]
##     screenWidget = _IdToWxObj[id]
##     count        = _IdToUIObj[id].widgets.index(screenWidget)
##     self._eventProcessor(GFEvent('requestFOCUS',gfObject))
##     self._eventProcessor(GFEvent('requestJUMPRECORD',count - gfObject._visibleIndex))
##     self._eventProcessor(GFEvent('requestCURSORMOVE',position=cursorPosition))

##     event.Skip()

class keyboardEvtHandler(uiBaseEvtHandler):
  def __init__(self, eventProcessor):
    pass
##     eventList = [wxEVT_CHAR,wxEVT_KEY_DOWN]
##     uiBaseEvtHandler.__init__(self, eventList, eventProcessor)
    
##   def _processEvent(self,event):
##     action = None
##     # hack for swig shortcomming needed by wxPython 2.2.x (2.3 does not need this)
##     # compute the location of the character in the widget
##     if event.GetEventType() == wxEVT_CHAR or event.GetEventType()==wxEVT_KEY_DOWN:
##       object = _eventObjTowxWindow(event)
##       command = GFKeyMapper.KeyMapper.getEvent(
##         event.KeyCode(),
##         event.ShiftDown(),
##         event.ControlDown(),
##         event.AltDown())

##       # TODO : Broken for the moment, this should probably be form driven 
##       #if command == 'JUMPRECORD':
##       #  self.promptForRecordNumber()

##       if command:
##         action = GFEvent('request%s' % command)

##       else:
##         try:
##           action = GFEvent('requestKEYPRESS', chr(event.KeyCode()),
##                            text=chr(event.KeyCode()), 
##                            code=event.KeyCode())
##         except ValueError:
##           pass #event.Skip()
        
##     if action:
##       self._eventProcessor(action)

#####################################################################
##
## Login Support
##
#####################################################################

#
# UILoginHandler
#
class UILoginHandler(GLoginHandler.LoginHandler):
  pass
##   def __init__(self):
##     self._wxapp = getWxApp()
##     self.dlg = None
    
##   def getLogin(self, loginData):
##     if 1: # TODO: This is left here to prevent me from having to unindent this code
##           # TODO: Since the UI stuff is currently being gutted I'm not wasting the time
##           # TODO: to do the right thing
##       if len(loginData[1]):
##         loginMesg = 'Login required for "%s"' % (loginData[1])
##       else:
##         loginMesg = 'Login required for %s' % (loginData[0])

##       self.dlg = wxDialog(NULL, -1, "GNU Enterprise: Login")
##       self.dlg.SetAutoLayout(true)
    
##       bmp = wxImage(os.environ['INSTALL_PREFIX']+'/'+gConfig('smallPNG'), wxBITMAP_TYPE_PNG).ConvertToBitmap()

##       self.textctrlList = []
##       messageField = wxStaticText(self.dlg, 1010, str(loginMesg))
##       labelList = []

##       dlgWidth = max(bmp.GetWidth(), messageField.GetSize().GetWidth()+20)

##       dlgHeight = bmp.GetHeight() + messageField.GetSize().GetHeight() + 80

##       xSpacing = 0
##       ySpacing = 0
##       fieldLabelWidth = 0

##       for prompt in loginData[2]: 
##         s = wxStaticText(self.dlg, 1010, '%s:' % prompt[1])
##         labelList.append(s)
##         if prompt[2]: 
##           t = wxTextCtrl(self.dlg, -1,"",wxPoint(1, 1), wxSize(150, 20), style=wxTE_PASSWORD|wxTE_PROCESS_ENTER)
##         else:
##           t = wxTextCtrl(self.dlg, -1,"",wxPoint(1, 1), wxSize(150, 20), style=wxTE_PROCESS_ENTER)
          
##         myID = len(self.textctrlList)
##         self.textctrlList.append(t)
##         EVT_CHAR(t, LoginFieldHandler(self, myID).loginFieldEventTrap)

##         fieldLabelWidth = max(fieldLabelWidth, \
##                                      s.GetSize().GetWidth() + t.GetSize().GetWidth() + 10)   

##         dlgWidth = max(dlgWidth, \
##                               s.GetSize().GetWidth() + t.GetSize().GetWidth() + 10)

##         xSpacing = max(xSpacing, s.GetSize().GetWidth())
##         ySpacing = max(ySpacing, s.GetSize().GetHeight())
##         ySpacing = max(ySpacing, t.GetSize().GetHeight())


##       loginButton = wxButton(self.dlg,19998,'Login')
##       cancelButton = wxButton(self.dlg,19999,'Cancel')

##       EVT_BUTTON(self.dlg, 19998, self.loginButtonEventTrap)
##       EVT_BUTTON(self.dlg, 19999, self.loginCancelEventTrap)                
##       dlgWidth = max(dlgWidth, loginButton.GetSize().GetWidth() + 
##                             cancelButton.GetSize().GetWidth() + 6) + 20

##       dlgHeight += max(loginButton.GetSize().GetHeight(), 
##                        cancelButton.GetSize().GetHeight()) - 6

##       firstY = bmp.GetHeight() + messageField.GetSize().GetHeight() + 50
##       lastY = firstY
##       xSpacing += 10	# Add whitespace between widgets
##       ySpacing += 6	# Add whitespace between widgets
##       xPos = dlgWidth/2 - fieldLabelWidth/2
      

##       # Move the fields and labels into position
##       for i in range(0, len(self.textctrlList)):
##         dlgHeight = dlgHeight + ySpacing
##         labelList[i].SetPosition(wxPoint(xPos, lastY))
##         self.textctrlList[i].SetPosition(wxPoint(xPos + xSpacing, lastY))
##         lastY = lastY + ySpacing

##       # Set the focus to the first text entry field
##       self.textctrlList[0].SetFocus()

##       # Create and position the logo
##       wxStaticBitmap(self.dlg,-1, bmp,
##                      wxPoint((dlgWidth-bmp.GetWidth())/2, 12),
##                      wxSize(bmp.GetWidth(), bmp.GetHeight()))
      
##       # Move the various widgets into position
##       messageField.SetPosition(
##         wxPoint(dlgWidth/2 - messageField.GetSize().GetWidth()/2,
##                 30 + bmp.GetHeight()))
      
##       cancelButton.SetPosition(
##         wxPoint(dlgWidth - 10 - cancelButton.GetSize().GetWidth(),
##                 dlgHeight - 10 - max(loginButton.GetSize().GetHeight(),
##                                             cancelButton.GetSize().GetHeight())))
##       loginButton.SetPosition(
##         wxPoint(dlgWidth - 16 - cancelButton.GetSize().GetWidth() - \
##                 loginButton.GetSize().GetWidth(),
##                 dlgHeight - 10 - max(loginButton.GetSize().GetHeight(),
##                                             cancelButton.GetSize().GetHeight())))

##       self.loginButton = loginButton

##       self.dlg.SetSize(wxSize(dlgWidth, dlgHeight))


##       self.dlg.Refresh()
##       self.dlg.Fit()
##       self.dlg.Raise()
##       self.dlg.CenterOnScreen()


##     # If user cancels, this will be set to 0
##     self._completed = 0
##     self.dlg.ShowModal()
      
##     if not self._completed:
##       raise GLoginHandler.UserCanceledLogin

##     rv = {}
##     for i in range(0, len(loginData[2])):
##       rv[loginData[2][i][0]] = self.textctrlList[i].GetValue()

##     return rv

##   #
##   # Login is completed, for whatever reason
##   #
##   def loginCompleted(self, successful):
##     self._completed = successful
##     self.dlg.EndModal(1)

##   #
##   # Called when user clicks "login"
##   #
##   def loginButtonEventTrap(self, event):
##     self.loginCompleted(1)

##   #
##   # Called when user clicks "cancel"
##   #
##   def loginCancelEventTrap(self, event):
##     self.loginCompleted(0)

##   #
##   # TODO: This is a hack required because windows
##   # TODO: seems to have issues with wxWindows dialogs
##   #
##   def destroyLoginDialog(self):
##     self.dlg.Destroy()

## #
## # LoginFieldHandler
## #
## # Used by the login handler
## # enables the user to press return and have it jump to the next box
## #
## class LoginFieldHandler:
##   def __init__(self, app, seq):
##     self.app = app
##     self.seq = seq

##   def loginFieldEventTrap(self, event):
##      if event.KeyCode() in (WXK_RETURN, WXK_TAB):
##        if self.seq < len(self.app.textctrlList) - 1:
##          self.app.textctrlList[self.seq+1].SetFocus()
##        else:
##          if event.KeyCode() == WXK_TAB:
##            if event.ShiftDown():
##              self.app.textctrlList[self.seq-1].SetFocus()
##            else:
##              self.app.loginButton.SetFocus()
##          else:
##            self.app.loginCompleted(1)
##      else:
##       event.Skip()








